package ZipTie::Adapters::Brocade::vRouter::Parsers;

use strict;
use warnings;
use 5.010_000;

use MIME::Base64 qw(encode_base64);
use ZipTie::Adapters::Utils qw(strip_mac seconds_since_epoch get_mask get_port_number get_interface_type get_crep);
use ZipTie::Addressing::Subnet;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_interfaces print_interfaces create_config parse_chassis parse_filters parse_system parse_static_routes parse_object_groups print_object_groups);

my $LOGGER = ZipTie::Logger::get_logger();
my %interface_subnets;
my %network_groups;
my %service_groups;

sub parse_system {
	my ( $in, $out ) = @_;

	if( $in->{showName} =~ /show\s+host\s+name.+?^\s*(\S+)/sm ) {
		$out->print_element( 'core:systemName', $1);
	}

	if( $in->{showVer} =~ /Version\s*:\s+([\w\.]+)/i ) {
		$out->open_element('core:osInfo');
		$out->print_element('core:make', 'Brocade');
		$out->print_element('core:version', $1);
		$out->print_element('core:osType', 'vRouter');
		$out->close_element('core:osInfo');
	}

	my $device_type = 'Router';
	if( $in->{config} =~ /^(\s{0,6})firewall\s*{(.+?)^\1}/msi ) {
		$device_type = 'Firewall';
	}
	$out->print_element( 'core:deviceType', $device_type );

	# System booted: Tue Apr 17 19:23:10 UTC 2007
	my $last_reboot = -1;
	if( $in->{showVer} =~ /System booted:\s+\S+\s+(\S+)\s+(\d+)\s+(\d{1,2}:\d{1,2}:\d{1,2})\s+(\S+)\s+(\d{4})/i ) {
		my ( $month, $day, $time, $timezone, $year ) = ( $1, $2, $3, $4, $5 );
		my ( $hour, $min, $sec ) = $time =~ /(\d+):(\d+):(\d+)/;
		$last_reboot = seconds_since_epoch( $sec, $min, $hour, $day, $month, $year, $timezone );
	}
	$out->print_element( 'core:lastReboot', $last_reboot );

	return;
}

sub parse_chassis {
	my ( $in, $out ) = @_;

	my $MAKE = 'Brocade';
	my $MODEL = 'vRouter';
	my $chassis = {
		'core:asset' => {
			'core:assetType' => 'Chassis',
			'core:factoryinfo' => {
				'core:make' => $MAKE,
				'core:modelNumber' => $MODEL
			}
		},
		'core:description' => $MODEL
	};

	$in->{showVer} =~ s/#.+//g; # remove comments in the output
	if( $in->{showVer} =~ /HW\s+S\/N:\s*(.+)\s*/i ) {
		my $serial_number = $1;
		$serial_number =~ s/\s+$//g; # remove trailing spaces, seen on customer site
		$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $serial_number;
	}

	if( $in->{showMem} =~ /Total:\s*(\d*)/i ) {
		$chassis->{memory}->{'core:description'} = 'RAM';
		$chassis->{memory}->{kind}               = 'RAM';
		$chassis->{memory}->{size}               = int $1;
	}

	$out->print_element( 'chassis', $chassis );
	return;
}

sub parse_static_routes {
	my ( $in, $out ) = @_;
	my $ip_regex = get_crep('cipm');
	my @static_routes;
	my $found_default_gateway;

	if( $in->{config} =~ /^\s{1,12}gateway-address\s+($ip_regex)\s*/mi ) {
		my $route = {
			destinationAddress => '0.0.0.0',
			destinationMask => '0',
			gatewayAddress => $1,
			defaultGateway => 'true',
			routeMetric => '1',
			routePreference => '1'
		};
		push @static_routes, $route;
		$found_default_gateway = 1;
	}
	if( $in->{config} =~ /^(\s{1,6})static\s*{(.+?)^\1}/msi ) {
		my $blob = $2;
		while( $blob =~ /^(\s+)route\s+($ip_regex)\/(\d+)\s+{(.+?)^\1}/msgi ) {
			my $route = {
				destinationAddress => $2,
				destinationMask    => $3
			};
			if( $4 =~ /^(\s+)next-hop\s+($ip_regex)\s+{(.+?)^\1}/msi ) {
				$route->{gatewayAddress} = $2;
				$route->{routePreference} = '1';
				$route->{routeMetric} = '1';
				if( $3 =~ /^\s+distance\s+(\d+).*/msi ) {
					$route->{routeMetric} = $1; 
				}
				my $is_default_gateway = 'false';
				if( not $found_default_gateway and $route->{destinationAddress} eq '0.0.0.0' and $route->{destinationMask} eq '0' ) {
					$is_default_gateway = 'true';
				}
				$route->{defaultGateway} = $is_default_gateway;

				push @static_routes, $route ;
			}
		}
	}

	# Need to find the interface to take to the gateway, drop any we cannot resolve
	my @valid_static_routes;
	for my $static_route ( @static_routes ) {
		my $found = 0;

		INTERFACE_SEARCH:
		for my $interface ( keys %interface_subnets ) {
			for my $subnet ( @{ $interface_subnets{$interface} } ) {
				if( $subnet->contains( $static_route->{gatewayAddress} ) ) {
					$static_route->{interface} = $interface;
					$found = 1;
					last INTERFACE_SEARCH;
				}
			}
		}
		if( $found ) {
			push @valid_static_routes, $static_route;
		} else {
			$LOGGER->debug("WARNING: Could not resolve interface for gateway address '$static_route->{gatewayAddress}', dropping route.");
		}
	}
	@static_routes = @valid_static_routes;

	if( @static_routes ) {
		$out->print_element( 'staticRoutes', { staticRoute => \@static_routes } );
	}
	return;
}

sub parse_interfaces {
	my ( $in, $out ) = @_;
	my $ip_regex = get_crep('cipm');
	my $v6_regex = get_crep('ipv6');
	my ($interface_config) = _get_section( 'interfaces', $in->{config} );

	my @interfaces;

	my $header_printed = 0;
	while( $interface_config =~ /^(\s+)(\S+)\s+(\S+)\s*{(.+?)^\1}/msg ) {
		my ($type, $name, $blob) = ($2, $3, $4);

		my $interface = {
			name          => $name,
			interfaceType => get_interface_type($type),
			physical      => _is_physical($name),
		};
		
		if( $blob =~ /^\s*description\s+"?(.+)"?/mi ) {
			$interface->{description} = $1;
		}

		if( $blob =~ /^\s*hw-id\s+(\S+)/mi ) {
			$interface->{interfaceEthernet}->{macAddress} = strip_mac($1);
		}

		if( $name =~ /eth/i ) {
			$interface->{interfaceEthernet}->{autoSpeed}  = ( $blob =~ /^\s+speed\s+\"?auto\"?/mi )  ? "true" : "false";
			$interface->{interfaceEthernet}->{autoDuplex} = ( $blob =~ /^\s+duplex\s+\"?auto\"?/mi ) ? "true" : "false";
		}

		my @ip_configurations;
		while( $blob =~ /^(\s*)address\s+($ip_regex)\/(\d+)\s*$/msig ) {
			my $ip_configuration = { ipAddress => $2, mask => $3 };
			push @ip_configurations, $ip_configuration;
		}
		while( $blob =~ /^(\s*)address\s+($ip_regex)\s+{(.+?)^\1}/msig ) {
			my $props = $3;
			my $ip_configuration = { ipAddress => $2, };
			if( $props =~ /^\s*prefix-length:\s*(\d+)/mi ) {
				$ip_configuration->{mask} = $1;
			}
			push @ip_configurations, $ip_configuration;
		}
		$interface->{interfaceIp}->{ipConfiguration} = \@ip_configurations;

		# Store for later lookup by route parsing
		for my $ip_configuration ( @ip_configurations ) {
			push @{ $interface_subnets{$name} }, ZipTie::Addressing::Subnet->new( $ip_configuration->{ipAddress}, $ip_configuration->{mask} ); 
		}

		while( $blob =~ /^(\s*)address\s+($v6_regex)\/(\d+)\s*$/msig ) {
			my ($address, $mask) = ($2, $3);
			$LOGGER->debug("WARNING: IPv6 addresses are not supported by QRM. Dropping $address/$mask from interface $name");
		}

		my ($firewall) = _get_section( 'firewall', $blob );
		if( $firewall ) {
			if( $firewall =~ /\s+local\s+({\s+name\s+)?"?(.+?)"?\s*$/mi ) {
				$interface->{ingressFilter} = $2;
			}
			if( $firewall =~ /\s+in\s+({\s+name\s+)?"?(.+?)"?\s*$/mi ) {
				$interface->{ingressFilter} = $2;
			}
			if( $firewall =~ /\s+out\s+({\s+name\s+)?"?(.+?)"?\s*$/mi ) {
				$interface->{egressFilter} = $2;
			}
		}

		my $status = 'up';
		if( $blob =~ /^\s+disable:?\s+true/mi ) {
			$status = 'down';
		}
		if( $in->{interfaces} =~ /^\s*$name.+\s+([uDA])\/([uDA])/m ) {
			unless( $1 eq 'u' and $2 eq 'u' ) {
				$status = 'down';
			}
		}
		$interface->{adminStatus} = $status;

		push @interfaces, $interface;
	}

	$in->{parsed}->{interfaces} = { interface => \@interfaces };
	return;
}

sub print_interfaces {
	my ( $in, $out ) = @_;
	if( $in->{parsed}->{interfaces} ) {
		$out->print_element( 'interfaces', $in->{parsed}->{interfaces} );
	}
	return;

}

sub parse_object_groups {
	my ( $in, $out ) = @_;

 	
	if( $in->{config} =~ /^(\s+)group\s*{(.+?)^\1}/msi ) {
		my $blob = $2;

		my $network_groups = _parse_network_objects($blob);
		my $service_groups = _parse_service_objects($blob);
		$in->{parsed}->{object_groups} = { networkGroup => $network_groups, serviceGroup => $service_groups };
	}

	return;
}

sub print_object_groups {
	my ( $in, $out ) = @_;
	if( $in->{parsed}->{object_groups} ) {
		$out->open_element( 'objectGroups' );
		my $network_groups = $in->{parsed}->{object_groups}->{networkGroup};
		my @elements = qw( objectGroupReference host range subnet ); 
		for my $network_group ( @{ $network_groups } ) {
			$out->open_element( 'networkGroup' );
			$out->print_element( 'id', $network_group->{id} );
			for my $element ( @elements ) {
				if( defined $network_group->{$element} ) {
					$out->print_element( $element, $network_group->{$element} ); }
				}
			$out->close_element( 'networkGroup' );
		}
		if( $in->{parsed}->{object_groups}->{serviceGroup} ) {
			$out->print_element( 'serviceGroup', $in->{parsed}->{object_groups}->{serviceGroup} );
		}
		$out->close_element( 'objectGroups' );
	}
	return;
}

sub _parse_network_objects {
	my ( $config ) = @_;
	my $ip_regex = get_crep('cipm');
	my $network_groups;

	while( $config =~ /^(\s+)address-group\s+(\S+)\s+{(.+?)^\1}/msig ) {

		my ($id, $inner_blob) = ($2, $3);
		my $network_group = { id => $id };
		while( $inner_blob =~ /address\s+($ip_regex)(\/(\d+)|-($ip_regex))?/g ) {
			my ($address, $mask, $range_end) = ($1, $3, $4);

			if( $range_end ) {
				push @{ $network_group->{range} }, { startAddress => $address, endAddress => $range_end };
			} elsif( not defined $mask or $mask == 32 ) {
				push @{ $network_group->{host} }, $address;
			} else {
				push @{ $network_group->{subnet} }, { address => $address, mask => $mask };
			}
		}
		push @{ $network_groups }, $network_group;
		$network_groups{$id} = 1; # add ID to our hash for easy lookup later
	}

	return $network_groups;
}

sub _parse_service_objects {
	my ( $config ) = @_;
	my $service_groups;

	SERVICE_GROUP:
	while( $config =~ /^(\s+)port-group\s+(\S+)\s+{(.+?)^\1}/msig ) {
		my ($id, $inner_blob) = ($2, $3);
		my $service_group = { id => $id };

		while( $inner_blob =~ /port\s+([\x21-\x7E]+)/g ) {
			my $port = $1;

			# Is the port a number or a well known name? Try to resolve it either way
			if( $port !~ /^\d+$/ ) {
				my $decoded_port = get_port_number($port);
				if( $decoded_port =~ /^\d+$/ ) {
					$port = $decoded_port;
				} else {
					$LOGGER->debug("WARNING: Could not parse port '$port', dropping service object '$id'.");
					next SERVICE_GROUP;
				}
			}
			push @{ $service_group->{portExpression} }, { port => $port, operator => 'eq' };
		}
		push @{ $service_groups }, $service_group;
		$service_groups{$id} = 1;
	}

	return $service_groups;
}

# TODO: REMOVE. THERE IS OR SHOULD BE A COMMON VERSION OF THIS
# Populates the configuration entity for the main Brocade config
sub create_config {
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository;
	$repository->{"core:name"} = "/";

	# build the simple text configuration
	my $config;
	$config->{"core:name"}       = "config";
	$config->{"core:textBlob"}   = encode_base64( $in->{config} );
	$config->{"core:mediaType"}  = "text/plain";
	$config->{"core:context"}    = "N/A";
	$config->{'core:promotable'} = 'false';                          

	# push the configuration into the repository
	push( @{ $repository->{'core:config'} }, $config );

	# print the repository
	$out->print_element( "core:configRepository", $repository );
	return;
}

# put the brocade firewall rules into the ZipTie filters model
sub parse_filters {
	my ( $in, $out ) = @_;

	my $blob = $in->{config};
	if( $blob =~ /^(\s*)security\s*{(.+?)^\1}/msi) {
		$blob = $2;
	}
	if( $blob =~ /^(\s*)firewall\s*{(.+?)^\1}/msi ) {
		$blob = $2;
		$out->open_element("filterLists");

		# iterate over each firewall set
		while ( $blob =~ /^(\s+)name\s+(\b.+?\b)\s*{(.+?)^\1}/msig ) {
			my $process_order = 1;
			my $firewall_name = $2;
			my $rules         = $3;

			$out->open_element('filterList');

			if( $rules =~ /^\s+default-action\s+"?(\w+)"?/mi ) {
				my $action = $1 eq 'accept' ? 'permit' : 'drop';
				$out->print_element( 'defaultAction', $action );
			}

			if( $rules =~ /^\s+description\s*(")?([\x20-\x7E]+)\1/mi ) {
				$out->print_element( "description", $2 );
			}

			# iterate over each rule
			while( $rules =~ /^(\s+)rule\s+(\d+)\s*{(.+?)^\1}/msig ) {
				my ($rule_name, $rule_blob) = ($2, $3);
				my $rule = _parse_rule($rule_blob, $rule_name, $process_order);
				if( $rule ) {
					$out->print_element( "filterEntry", $rule );
					++$process_order;
				}
			}

			$out->print_element( "mode", 'stateful' );
			$out->print_element( "name", $firewall_name );
			$out->close_element("filterList");
		}
		$out->close_element("filterLists");
	}
	return;
}

sub _parse_rule {
	my ($config, $name, $process_order) = @_;

	my $rule;
	$rule->{name} = $rule->{ztSignature} = $name;
	$rule->{processOrder} = $process_order;

	# does this rule permit or deny
	$rule->{primaryAction} = $config =~ /^\s+action\s*\"?accept\"?/m ? 'permit' : 'drop';

	# when hit does this rule produce a log entry
	$rule->{log} = $config =~ /^\s+log/m ? 'true' : 'false';

	# applies to established connections only
	$rule->{established} = $config =~ /^\s+established\s+enable/m ? 'true' : 'false';

	my ($protocols, $error) = _parse_rule_protocol($config, $name);
	if( $error ) {
		$LOGGER->debug($error);
		return;
	}

	while( $config =~ /^(\s+)(source|destination)\s*{(.+?)^\1}/msig ) {
		my ($location, $blob) = ($2, $3);

		(my $address, $error) = _parse_rule_address($blob, $name);
		if( $error ) {
			$LOGGER->debug($error);
			return;
		}
		if( $address ) {
			$rule->{ $location . 'IpAddr' } = $address;
		}

		(my $service, $error) = _parse_rule_service($blob, $name, $protocols);
		if( $error ) {
			$LOGGER->debug($error);
			return;
		}
		if( $service ) {
			$rule->{ $location . 'Service' } = $service;
		}
	}

	# Set protocol field if no source or destination service was created and protocol exists in rule
	if( not exists $rule->{sourceService} and not exists $rule->{destinationService} and $protocols ) {
		my @protocols = @{ $protocols };
		if( scalar @protocols == 1 ) {
			$rule->{protocol} = $protocols[0];
		}
	}

	return $rule;
}

sub _parse_rule_protocol {
	my ($blob, $rule_name) = @_;
	my @VALID_PROTOCOLS = qw( tcp udp icmp );
	my $protocols;
	my $error;

	if( $blob =~ /^\s+protocol\s*\"?([\x21-\x7E]+)\"?/m ) {
		my $protocol = $1;
		if( $protocol eq 'all' ) {
			$protocols = \@VALID_PROTOCOLS;
		} else {
			for my $p ( @VALID_PROTOCOLS ) {
				if( $protocol =~ /$p/ ) {
					push @{ $protocols }, $p;
				}
			}
		}
		if( not $protocols ) {
			$error = "WARNING: Protocol '$protocol' not supported, dropping rule '$rule_name'";
		}
	}

	return ($protocols, $error);
}

sub _parse_rule_address {
	my ($blob, $rule_name) = @_;
	my $ip_regex = get_crep('cipm');
	my $v6_regex = get_crep('ipv6');
	my $parsed_address;
	my $error;

	# \x21-\x7E is all the common valid chars
	if( $blob =~ /address(-group)?\s+([\x21-\x7E]+)/ ) {
		my $address = $2;

		if( $address =~ /^$v6_regex(\/(\d+))?$/ ) {
			$error = "WARNING: Found IPv6 address '$address' in rule '$rule_name', dropping rule.";
		} elsif( $address =~ /^($ip_regex)(\/(\d+))?$/ ) {
			my ($ip, $mask) = ($1, $3);
			if( not defined $mask or $mask == 32 ) {
				$parsed_address = { host => $ip };
			} else {
				$parsed_address = { network => { address => $ip, mask => $mask } };
			}
		} elsif( $network_groups{$address} ) {
			$parsed_address = { objectGroupReference => $address };
		} else {
			$error = "WARNING: Could not resolve address '$address' in rule '$rule_name', dropping rule."
		}
	}

	return ($parsed_address, $error);
}

sub _parse_rule_service {
	my ($blob, $rule_name, $protocols) = @_;
	my $parsed_services;
	my $error;

	if( $blob =~ /port\s+([^!].+)/ ) {
		my @port_expressions = split /,/, $1;
		for my $port_expression ( @port_expressions ) {
			my $parsed_service;

			if( $port_expression =~ /^\s*(\d+)-(\d+)\s*$/ ) {
				for my $protocol ( @{ $protocols } ) {
					push @{ $parsed_services }, { portRange => { portStart => $1, portEnd => $2, protocol => $protocol } };
				}
			} elsif( $port_expression =~ /^\s*(\d+)\s*$/ ) {
				for my $protocol ( @{ $protocols } ) {
					push @{ $parsed_services }, { portExpression => { port => $1, operator => 'eq', protocol => $protocol } };
				}
			} elsif( $port_expression =~ /^\s*([\x21-\x7E]+)/ ) {
				my $port_name = $1;
				my $port_number = get_port_number($port_name);
				if( $port_number =~ /^\d+$/ ) {
					for my $protocol ( @{ $protocols } ) {
						push @{ $parsed_services }, { portExpression => { port => $port_number, operator => 'eq', protocol => $protocol } };
					}
				} elsif( $service_groups{$port_name} ) {
					push @{ $parsed_services }, { objectGroupReference => $port_name };
				} else {
					$error = "WARNING: Could not resolve service '$port_name' in rule '$rule_name', dropping rule.";
					return (undef, $error);
				}
			}
		}
	} elsif ( $blob =~ /port\s+!(.+)/ ) {
		$error = "WARNING: Negating a service object is not supported, dropping rule '$rule_name'";
	}

	return ($parsed_services, $error);
}

# returns the 0.0.0.0/0 address
sub _get_any {
	return { address => '0.0.0.0', mask => '0' };
}

# given the name of a configuration section, returns an array of each section
sub _get_section {
	my ( $section, $input ) = @_;
	my @results;
	while( $input =~ /^((\s*)$section[^\n]*?{(.+?)^\2})/msig ) {
		push @results, $1;
	}
	return @results;
}

sub _is_physical {
	my $name = shift;
	if( $name =~ /\.\d+$/ ) {
		return 'false';
	} elsif( $name =~ /seri|eth|gig|^fe|^fa|token|[bp]ri/i ) {
		return 'true';
	} else {
		return 'false';
	}
}

1;
