package ZipTie::Adapters::CheckPoint::HTTPS;

use strict;
use warnings;
use 5.010_000;

use lib '/opt/qradar/lib/perl';
use JSON;

use ZipTie::Adapters::CheckPoint::HTTPS::Parsers qw(parse_system parse_chassis parse_interfaces parse_object_groups parse_filters parse_static_routing);
use ZipTie::Adapters::CheckPoint::HTTPS::Utils qw(parse_sid get_server_by_ip get_response_objects remote_run_script);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);

use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::RESTAPI;

my $LOGGER = ZipTie::Logger::get_logger();

sub backup {
	my ($package, $backup_xml) = @_;

	# Turn the backup operation XML document into ZipTie::ConnectionPath
	my ($connection_path) = ZipTie::Typer::translate_document($backup_xml, 'connectionPath');
	my $responses = get_responses($connection_path);

	# Set up model printing object
	my $filehandle = get_model_filehandle('Check Point HTTPS', $connection_path->get_ip_address());
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'common');
	$printer->open_model();

	# Parse and print model
	parse_system($responses, $printer);
	parse_chassis($responses, $printer);
	parse_object_groups($responses, $printer);
	parse_filters($responses, $printer);
	parse_interfaces($responses, $printer);
	parse_static_routing($responses, $printer);

	$printer->close_model();
	close_model_filehandle($filehandle);

	return;
}

sub get_responses {
	my ($connection_path) = @_;
	my $responses;

	my $mgmt_ip = $connection_path->get_management_server_ip_address();
	my $ip = $connection_path->get_ip_address();
	$responses->{ip} = $ip;

	# Create RESTAPI object for communicating with the Management API
	my $connection = ZipTie::RESTAPI->new($connection_path, '/web_api/');

	# Set header declaring we are sending JSON
	$connection->{headers} = [ 'Content-Type' => 'application/json' ];

	# Authentication and get the Session ID
	my $credentials = $connection_path->get_credentials();
	my $authentication_body = {
		user => $credentials->{username},
		password => $credentials->{password}
	};
	$LOGGER->debug("Using username: $credentials->{username}");
	if( $credentials->{enableUsername} ) {
		$authentication_body->{domain} = $credentials->{enableUsername};
		$LOGGER->debug("Using domain  : $credentials->{enableUsername}");
	}
	$authentication_body = encode_json($authentication_body);

	# Store the Session ID in a header to be passed each consequent call
	my $sid = $connection->obtain_token_as_header(
		'login',                                # API endpoint
		$authentication_body,                   # POST body (encoded to JSON)
		\&parse_sid,                            # Sub that parses SID from response
		'x-chkp-sid'                            # Header name for SID
	);

	# Ordered list of API endpoint calls
	my @commands = (
		'show-gateways-and-servers',
		'show-hosts',
		'show-networks',
		'show-address-ranges',
		'show-groups',
		'show-groups-with-exclusion',
		'show-services-tcp',
		'show-services-udp',
		'show-service-groups',
		'show-packages'
	);

	# Call the API for each endpoint
	for my $command ( @commands ) {
		$responses->{$command} = get_response_objects($connection, $command);
	}

	# Callbacks for API queries we do not know the parameters of until parsing has started
	$responses->{'show-access-rulebase'} = sub {
		my $rulebase = shift;
		my $body = { name => $rulebase, 'details-level' => 'full', 'use-object-dictionary' => 'false' };
		my $response = $connection->get("show access rulebase $rulebase", 'show-access-rulebase', encode_json($body));
		return decode_json($response);
	};

	$responses->{'show-nat-rulebase'} = sub {
		my $package = shift;
		my $body = { package => $package, 'details-level' => 'full', 'use-object-dictionary' => 'false' };
		my $response = $connection->get("show nat rulebase $package", 'show-nat-rulebase', encode_json($body));
		return decode_json($response);
	};

	# Not all configuration is available directly from the mgmt API, such as the hostname
	# Before running any calls to run-script we need to find the target management server
	my $mgmt_server = get_server_by_ip($responses, $mgmt_ip);

	$responses->{hostname} = remote_run_script($connection, $mgmt_server, $ip, 'hostname');
	$responses->{ip_address} = remote_run_script($connection, $mgmt_server, $ip, 'ip address');
	$responses->{route} = remote_run_script($connection, $mgmt_server, $ip, 'route -n');

	return $responses;
}

1;
