
# Purp: Populate ZTS with managed Checkpoint devices
# Date: 14/07/2016

#!/usr/bin/perl

use strict;
use warnings;

use lib '/opt/qradar/lib/perl';
use JSON;

use ZipTie::Logger;
use ZipTie::Client;
use ZipTie::Adapters::Utils qw{get_crep};
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::Credentials;
use ZipTie::Recording;
use ZipTie::Adapters::CheckPoint::HTTPS::Utils qw(parse_sid get_response_objects);
use ZipTie::RESTAPI;

$ENV{'PERL_LWP_SSL_VERIFY_HOSTNAME'} = 0;

if(scalar(@ARGV) != 1){
 	_fatal("Wrong number of arguments passed to httpsdiscovery.pl");
}
my $ui_ip = $ARGV[0];

# Start logging
my $LOGGER = ZipTie::Logger::get_logger();
$LOGGER->debug("$ui_ip");
main($ui_ip); 

sub main{
	my ($management_server_ip) = @_;

	# Check parameter is a valid ip
	my $ip_regex = get_crep("cipm");
	if($management_server_ip !~ /^$ip_regex$/){
		_fatal("Invalid IP address: $management_server_ip");
	}

	# Get credentials for management device
	my $ziptie_client = ZipTie::Client->new();	
	my ($username, $password, $port) = _get_credentials($ziptie_client, $management_server_ip);

	# Connect to the management device and get managed device IPs, also get a hash of any pre existing devices in ZTS
	$LOGGER->debug("Connecting to Checkpoint with username $username and port $port");
	my $found_devices = _get_managed_devices($management_server_ip, $username, $password, $port);
	$LOGGER->debug("Found " . scalar(@$found_devices) . " device(s) on Checkpoint HTTPS");

	$LOGGER->debug("Collecting devices already in CSM");
	my $existing_devices = _get_devices($ziptie_client);
	$LOGGER->debug("Found ". scalar(keys(%{$existing_devices})) . " device(s) in CSM");

	my $device_service = $ziptie_client->devices();

	my $adapter = 'ZipTie::Adapters::CheckPoint::HTTPS';

	$LOGGER->debug("Discovering $adapter devices.");
	foreach my $found_device(@$found_devices){
		$LOGGER->debug("Found device: ". $found_device );
		_add_or_update_device($existing_devices, $device_service, $found_device, $adapter, $ui_ip);
	}
}

########################################################################
# Get username, password and port for connecting to the management blade 
########################################################################
sub _get_credentials{
	my ($ziptie_client, $management_server_ip) = @_;
	
	# Get credentials, fail if none are found
	my $credentials_service = $ziptie_client->credentials();
	my @credentials = $credentials_service->getCredentialSetsByIpAddress(ipAddress => $management_server_ip);
	if(!scalar(@credentials)){
		_fatal("No credentials found.");
	}

	# Get username and password for the management ip, fail if none are found
	my $username;
	my $password;
	if(ref($credentials[0]->{credentials}) ne "ARRAY"){
		$credentials[0]->{credentials} = [$credentials[0]->{credentials}];
	}
	foreach my $credential(@{$credentials[0]->{credentials}}){
		if($credential->{name} eq "username"){
			$username = $credential->{value};
		}elsif ($credential->{name} eq "password"){
			$password = $credential->{value};
		}
	}
	if(!$username || $username eq ''){
		_fatal("Authentication refused with an empty username.");
	}
	if(!$password || $password eq ''){
		_fatal("Authentication refused with an empty password.");
	}

	# Get port for HTTPS, if it is enabled, else fail
	my @protocols = $credentials_service->getProtocolConfigByIpAddress(ipAddress => $management_server_ip);
	if(!scalar(@protocols)){
		_fatal("No protocols found.");
	}
	my $port;
	foreach my $protocol (@{$protocols[0]->{protocols}}){
		if($protocol->{name} eq "HTTPS" && $protocol->{enabled} eq "true"){
			$port = $protocol->{port};
			last;
		}
	}

	if(!$port || $port eq ''){
		_fatal("Required protocol either disabled or undefined.");
	}

	return($username, $password, $port);
}

#######################################################################
# If the device is already added update it, otherwise add a new device
#######################################################################
sub _add_or_update_device{
	my ($existing_devices, $device_service, $device_ip, $adapter, $ui_ip) = @_;
	$LOGGER->debug("Analyzing device $device_ip...");

	my $device = $existing_devices->{$device_ip};
	
	# If device details have changed then update device, if it already exists then skip or else if it doesn't exist than add device
	if ($existing_devices->{$device_ip}){
		if (not $device->{managementServerIp} or $device->{managementServerIp} ne $ui_ip){
			$device->{managementServerIp} = $ui_ip;
			$device_service->updateDevice(ipAddress => $device_ip, managedNetwork => 'Default', device => $device);
			$LOGGER->debug("Updated $device_ip node.");
			print("$device_ip, Node updated.\n");
		} else {
			$LOGGER->debug("Skipping existing device: $device_ip.");
			print("$device_ip,Skipped\n");
		}
	}else{
		my %device = (ipAddress => $device_ip, adapterId => $adapter, managedNetwork => "Default", managementServerIp => $ui_ip);
		$device_service->createManagedDevice(%device);
		$existing_devices->{$device_ip} = \%device;
		$LOGGER->debug("Adding new device $device_ip.");
		print("$device_ip,Added\n");
	}
}

###########################################################################
# Get devices currently in CSM. Returns a hash where keys are devices' ips
###########################################################################
sub _get_devices{
	my ($ziptie_client) = @_;
	my %device_hash;

	# Variables for getting page of devices
	my $PAGE_SIZE = 500;
	my $offset = 0;
	my $total;
	my $device_search_service = $ziptie_client->devicesearch();
	
	# While there are still devices to get
	while(!defined($total) || $offset < $total){

		# Get devices, 500 at a time from the current offset
		my $page = $device_search_service->search(scheme => "ipAddress", query => '',	pageData => {pageSize => $PAGE_SIZE, offset => $offset});
		$offset += $PAGE_SIZE;
		$total = $total // $page->{total};

		# Get an array reference of the devices
		my $devices;
		if($page->{devices}){
			$devices = $page->{devices};
			if(ref($devices) ne "ARRAY"){
				$devices = [$devices];
			}
		}

		# For each device in the array add it to the hash with its ip as the key
		foreach my $device(@$devices){
			$device_hash{$device->{ipAddress}} = $device;
		}
	}
	return(\%device_hash);
}

##########################################################################
# Logs into the management blade and returns the ips of the managed devices
##########################################################################
sub _get_managed_devices{
	my ($management_server_ip, $username, $password, $port) = @_;

	# Create a connection path to the management blade 
	my $connection_path = ZipTie::ConnectionPath->new();
	$connection_path->set_management_server_ip_address($management_server_ip);
	$connection_path->set_credentials(ZipTie::Credentials->new(username => $username, password => $password));
	$connection_path->set_protocols([ZipTie::ConnectionPath::Protocol->new({name => "HTTPS", port => $port})]);
	
	
	# Connect to HTTPS
	my $connection = ZipTie::RESTAPI->new($connection_path, '/web_api/');

	# Set header declaring we are sending JSON
	$connection->{headers} = [ 'Content-Type' => 'application/json' ];

	my $authentication_body = {
                user => $username,
                password => $password
        };
	my $sid = $connection->obtain_token_as_header(
                'login',                                # API endpoint
                encode_json($authentication_body),      # POST body (encoded to JSON)
                \&parse_sid,                            # Sub that parses SID from response
                'x-chkp-sid'                            # Header name for SID
        );

	# Run the command to look in the file with the managed devices
	my $objects = get_response_objects($connection, 'show-gateways-and-servers');

	# Parse through the returned file and return an array of devices found, that we care about
	my @managed_devices;

	for my $obj ( @{ $objects } ) {

		my $device = $obj->{'ipv4-address'};
		my $netsec = $obj->{'network-security-blades'};

		if ($netsec->{"firewall"}) {
			push @managed_devices, $device;
		}
	}
	
	$LOGGER->debug("Managed devices: @managed_devices ");

    return(\@managed_devices);
}

#############################
# Used to print fatal errors 
#############################
sub _fatal{
	my ($message) = @_;
	print("ERROR: $message\n");
	$LOGGER->fatal($message);
	exit(1);
}

1;
