package ZipTie::Adapters::F5::BigIP;

use strict;
use warnings;

use ZipTie::Adapters::F5::BigIP::AutoLogin qw(execute);
use ZipTie::Adapters::F5::BigIP::Parsers qw(parse_chassis parse_filters parse_system parse_interfaces parse_static_routes parse_vlans parse_filter_enabled parse_default_filter get_bigip_version);
use ZipTie::Adapters::F5::BigIP::TMSHParsers qw{parse_tmsh_system parse_tmsh_chassis parse_tmsh_interfaces parse_tmsh_static_routes parse_tmsh_vlans parse_tmsh_filter_enabled parse_tmsh_default_filter parse_tmsh_filters};
use ZipTie::Adapters::F5::BigIP::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::Model::XmlPrint;
use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use Exporter qw(import);
our @EXPORT_OK = qw(get_version);

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
use ZipTie::Adapters::BaseAdapter;
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

my $LOGGER = ZipTie::Logger::get_logger();

sub backup{
	my $package_name = shift;

	# Retrieve the operation XML document that contains all of the IP, protocol, credential, and file server information
	# that is needed to successfully backup a device.
	my $backup_doc = shift;

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path) = ZipTie::Typer::translate_document($backup_doc, 'connectionPath');

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my ($cli_protocol, $device_prompt_regex) = _connect($connection_path);

	# Grab an output filehandle for the model.  This usually points to STDOUT
	my $filehandle = get_model_filehandle('BigIP', $connection_path->get_ip_address());
	
	# Initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'common');
	$printer->open_model();

	my $responses;

	my $version = get_version($cli_protocol, $device_prompt_regex);
	if($version >= 11){
		my $was_enabled = _disable_paging( $cli_protocol, $device_prompt_regex );
		my $previous_display_threshold = _disable_display_threshold( $cli_protocol, $device_prompt_regex );
		$responses = _get_tmsh_responses($cli_protocol, $device_prompt_regex);
		$responses->{packetFilterEnabled} = parse_tmsh_filter_enabled($responses, $printer);
		$responses->{defaultFilterAction} = parse_tmsh_default_filter($responses, $printer);
		parse_tmsh_system($responses, $printer);
		parse_tmsh_chassis($responses, $printer);
		parse_tmsh_filters($responses, $printer);
		parse_tmsh_interfaces($responses, $printer);
		parse_tmsh_static_routes($responses, $printer);
		parse_tmsh_vlans($responses, $printer);
		_enable_paging( $cli_protocol, $device_prompt_regex ) if $was_enabled;
		_reset_display_threshold( $cli_protocol, $device_prompt_regex, $previous_display_threshold ) if $previous_display_threshold;
	}else{
		$responses =_get_responses($cli_protocol, $device_prompt_regex);
		$responses->{packetFilterEnabled} = parse_filter_enabled($responses, $printer);
		$responses->{defaultFilterAction} = parse_default_filter($responses, $printer);
		parse_system($responses, $printer);
		parse_chassis($responses, $printer);
		parse_filters($responses, $printer);
		parse_interfaces($responses, $printer);
		parse_static_routes($responses, $printer);
		parse_vlans($responses, $printer);
	}

	$printer->close_model();
	close_model_filehandle($filehandle);
	disconnect($cli_protocol);
}

sub _connect{
	my $connection_path = shift;
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the BigIP device
	my $device_prompt_regex = execute($cli_protocol, $connection_path);

	# Store the found prompt as "prompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name("prompt", $device_prompt_regex);

	# Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
	return ($cli_protocol, $device_prompt_regex);
}

sub get_version{
	my ($cli_protocol, $device_prompt_regex) = @_;

	my $bigpipe_response = $cli_protocol->send_and_wait_for('bigpipe version show', $device_prompt_regex);
	my $tmsh_response = $cli_protocol->send_and_wait_for('show sys version', $device_prompt_regex);

	if($bigpipe_response =~ /VERSION\s+((\d+)[\d\.]+)/i || $tmsh_response =~ /VERSION\s+((\d+)[\d\.]+)/i){
		$LOGGER->debug(qq{Version found: '$1'});
		return $2;
	}
	
	$LOGGER->fatal(qq{Could not determine BIG-IP version!\nThis is a requirement to know whether to send bigpipe or tmsh commands.});
}

sub _disable_paging {
	my ( $cli_protocol, $device_prompt_regex ) = @_;
	$LOGGER->debug('Disabling paging within TMOS shell ...');

	# Check if paging is currently enabled so we can revert to it
	my $response = $cli_protocol->send_and_wait_for('list cli preference pager', $device_prompt_regex);
	my $was_enabled = $response !~ /pager\s+disabled/;

	$cli_protocol->send_and_wait_for('modify cli preference pager disabled', $device_prompt_regex);

	return $was_enabled;
}

sub _enable_paging {
	my ( $cli_protocol, $device_prompt_regex ) = @_;
	$LOGGER->debug('Re-enabling paging in TMOS shell ...');
	$cli_protocol->send_and_wait_for('modify cli preference pager enabled', $device_prompt_regex);
	return;
}

sub _disable_display_threshold {
	my ( $cli_protocol, $device_prompt_regex ) = @_;
	$LOGGER->debug('Disabling display threshold within TMOS shell ...');

	# Check if paging is currently enabled so we can revert to it
	my $response = $cli_protocol->send_and_wait_for('list cli preference display-threshold', $device_prompt_regex);
	my ($previous_display_threshold) = $response =~ /display-threshold\s+(\d+)/;

	$cli_protocol->send_and_wait_for('modify cli preference display-threshold 0', $device_prompt_regex);

	return $previous_display_threshold;
}

sub _reset_display_threshold {
	my ( $cli_protocol, $device_prompt_regex, $previous_display_threshold ) = @_;
	$LOGGER->debug('Resetting display threshold within TMOS shell ...');
	$cli_protocol->send_and_wait_for("modify cli preference display-threshold $previous_display_threshold", $device_prompt_regex);
	return;
}

# Returns a hash of responses from the device
sub _get_responses{
	my $cli_protocol = shift;
	my $device_prompt_regex = shift;
	my $responses;
	$responses->{global} = $cli_protocol->send_and_wait_for("bigpipe global",$device_prompt_regex);
	$responses->{hostname} = $cli_protocol->send_and_wait_for("bigpipe system hostname",$device_prompt_regex);
	$responses->{platform} = $cli_protocol->send_and_wait_for("bigpipe platform",$device_prompt_regex);
	$responses->{uptime} = $cli_protocol->send_and_wait_for("uptime",$device_prompt_regex);
	$responses->{version}  = $cli_protocol->send_and_wait_for("bigpipe version show",$device_prompt_regex);
	$responses->{license} = $cli_protocol->send_and_wait_for("cat /config/bigip.license",$device_prompt_regex);
	$responses->{filter_enabled} = $cli_protocol->send_and_wait_for("bigpipe db packetfilter",$device_prompt_regex);
	$responses->{default_filter} = $cli_protocol->send_and_wait_for("bigpipe db packetfilter.defaultaction",$device_prompt_regex);
	$responses->{filters} = $cli_protocol->send_and_wait_for("bigpipe packet filter list",$device_prompt_regex);
	$responses->{nats} = $cli_protocol->send_and_wait_for("bigpipe nat list all",$device_prompt_regex);
	$responses->{vlans} = $cli_protocol->send_and_wait_for("bigpipe vlan show all",$device_prompt_regex);
	$responses->{vlan_groups_list} = $cli_protocol->send_and_wait_for("bigpipe vlangroup list all",$device_prompt_regex);
	$responses->{vlan_groups} = $cli_protocol->send_and_wait_for("bigpipe vlangroup",$device_prompt_regex);
	$responses->{interfaces} = $cli_protocol->send_and_wait_for("ip addr list",$device_prompt_regex);
	$responses->{physical_interfaces} = $cli_protocol->send_and_wait_for("bigpipe interface show all",$device_prompt_regex);
	$responses->{interface_speeds} = $cli_protocol->send_and_wait_for("bigpipe interface all media speed",$device_prompt_regex);
	$responses->{trunks} = $cli_protocol->send_and_wait_for(q{bigpipe trunk all interfaces},$device_prompt_regex);
	$responses->{static_routes} = $cli_protocol->send_and_wait_for("route -n",$device_prompt_regex);
	$responses->{static_routes_list} = $cli_protocol->send_and_wait_for("bigpipe route all list all",$device_prompt_regex);
	$responses->{mgmt} = $cli_protocol->send_and_wait_for("bigpipe mgmt show all",$device_prompt_regex);
	$responses->{mgmt_route} = $cli_protocol->send_and_wait_for("bigpipe mgmt route show all",$device_prompt_regex);
	$responses->{pools} = $cli_protocol->send_and_wait_for("bigpipe pool",$device_prompt_regex);
	$responses->{selfs} = $cli_protocol->send_and_wait_for("bigpipe self",$device_prompt_regex);
	$responses->{virtuals} = $cli_protocol->send_and_wait_for("bigpipe virtual list all",$device_prompt_regex);
	$responses->{snats} = $cli_protocol->send_and_wait_for("bigpipe snat list all",$device_prompt_regex);
	$responses->{snat_pools} = $cli_protocol->send_and_wait_for("bigpipe snatpool list all",$device_prompt_regex);
	$responses->{snats_any_protocol} = $cli_protocol->send_and_wait_for(q{b db snat.anyipprotocol},$device_prompt_regex);
	return $responses;
}

sub _get_tmsh_responses{
	my ($cli_protocol, $device_prompt_regex) = @_;
	my $responses;

	my %commands = (
		hostname           => 'list sys global-settings hostname',
		mgmt               => 'list sys management-ip',
		memory             => 'show sys memory',
		hardware           => 'show sys hardware',
		version            => 'show sys version',
		filter_enabled     => 'list sys db packetfilter',
		default_filter     => 'list sys db packetfilter.defaultaction',
		snats_any_protocol => 'list sys db snat.anyipprotocol',
		interfaces         => 'list net interface all-properties',
		trunks             => 'list net trunk',
		filters            => 'list net packet-filter',
		vlans              => 'list net vlan all-properties',
		vlans_show         => 'show net vlan',
		vlan_groups_list   => 'list net vlan-group all all-properties',
		vlan_groups_show   => 'show net vlan-group',
		virtuals           => 'list ltm virtual',
		nats               => 'list ltm nat',
		snat_pools         => 'list ltm snatpool',
		snats              => 'list ltm snat',
		static_routes_list => 'list net route',
		pools              => 'list ltm pool',
		selfs              => 'list net self',
		ipsec              => 'list net ipsec',
		tunnels            => 'list net tunnels'
	);

	for my $key ( keys %commands ) {
		my $command = $commands{$key};
		$responses->{$key} = $cli_protocol->send_and_wait_for( $command, $device_prompt_regex );
	}

	return $responses;
}

1;
