package ZipTie::Adapters::Fortinet::FortiOS;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::Fortinet::FortiOS::AutoLogin qw(get_prompt);
use ZipTie::Adapters::Fortinet::FortiOS::Parsers qw(get_hardware_nics parse_object_groups create_config parse_chassis parse_filters parse_system parse_interfaces parse_static_routes parse_vlans parse_subelements);

use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::CLIProtocol;
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

use Exporter qw(import);
our @EXPORT_OK = qw(clean_responses get_services_predefined_list);

use Readonly;
Readonly my $SHOW_FULL_CONF_WAIT_SECONDS => 600;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

sub backup {
	my ($packageName, $backupDoc) = @_;

	# Translate the backup operation XML document into ZipTie::ConnectionPath
	my ($connectionPath) = ZipTie::Typer::translate_document( $backupDoc, 'connectionPath' );

	# Make a Telnet or SSH connection and get configuration
	my ($cliProtocol, $promptRegex) = _connect($connectionPath);
	my $responses = _get_responses($cliProtocol, $promptRegex);
	_disconnect($cliProtocol);

	# Set up the XmlPrint object for printing the ZiptieElementDocument (ZED)
	my $filehandle = get_model_filehandle( 'Fortinet FortiOS', $connectionPath->get_ip_address() );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	parse_system( $responses, $printer );
	parse_chassis( $responses, $printer );
	create_config( $responses, $printer );
	parse_object_groups( $responses, $printer );
	parse_filters( $responses, $printer, $responses->{mgmt_vdom} );
	parse_interfaces( $responses, $printer, $responses->{mgmt_vdom} );
	parse_static_routes( $responses, $printer );
	parse_subelements( $responses, $printer );

	parse_vlans( $responses, $printer );
	
	$printer->close_model();                # close out the ZiptieElementDocument
	close_model_filehandle($filehandle);    # Make sure to close the model file handle

	return;
}

sub _connect {
	my $connectionPath = shift;

	# Create a new CLI protocol object by using the ZipTie::CLIProtocolFactory::create sub-routine
	# to examine the ZipTie::ConnectionPath argument for any command line interface (CLI) protocols
	# that may be specified.
	my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);

	# Make a connection to and successfully authenticate with the device
	my $devicePromptRegex = ZipTie::Adapters::Fortinet::FortiOS::AutoLogin::execute( $cliProtocol, $connectionPath );

	# Store the regular expression that matches the primary prompt of the device under the key "prompt"
	# on the ZipTie::CLIProtocol object
	$cliProtocol->set_prompt_by_name( 'prompt', $devicePromptRegex );

	# Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
	return ( $cliProtocol, $devicePromptRegex );
}

sub _disconnect {
	my $cliProtocol = shift;

	# Close this session and exit
	$cliProtocol->send('exit');

	return;
}

sub _get_responses {
	my ($cli_protocol, $prompt_regex) = @_;
	my %responses;

	# Get VDOM status, the cli changes depending if its active or not
	$responses{status} = $cli_protocol->send_and_wait_for( 'get system status', $prompt_regex );
	my $vdoms_enabled = $responses{status} =~ /^Virtual\sdomain\sconfiguration:\senable\s*$/m;
	
	# If VDOMs are enabled need to switch to global mode to disable paging
	my $global_prompt_regex = $vdoms_enabled ? _switch_to_global($cli_protocol) : $prompt_regex;
	my $re_enable_paging = _disable_pagination($cli_protocol, $global_prompt_regex);
	$cli_protocol->send_and_wait_for( 'end', $prompt_regex ) if $vdoms_enabled;

	# Get the full configuration
	$responses{config} = $cli_protocol->send_and_wait_for( 'show full-configuration', $prompt_regex, $SHOW_FULL_CONF_WAIT_SECONDS );

	# Get management VDOM name
	my ($mgmt_vdom_name) = $responses{config} =~ /^\s*set\smanagement-vdom\s+"(\S+)"\s*$/m;
	$responses{mgmt_vdom} = $mgmt_vdom_name;

	# On VDOM enabled systems switch to "config global" before running certain commands
	# get the new prompt or use the existing one for non-vdom devices
	_switch_to_global($cli_protocol) if $vdoms_enabled;
	my $global_responses = _get_global_responses( $cli_protocol, $global_prompt_regex );
	%responses = ( %responses, %{ $global_responses } );

	# On VDOM enabled systems switch to the management vdom before running certain commands
	my $mgmt_prompt_regex = $prompt_regex;
	if( $vdoms_enabled ) {
		$cli_protocol->send_and_wait_for( 'end', $prompt_regex );
		$cli_protocol->send_and_wait_for( 'config vdom', '\(vdom\)\s+#\s*$' );
		$cli_protocol->send_and_wait_for( "edit $mgmt_vdom_name", "\\($mgmt_vdom_name\\)\\s+#\\s*\$" );
		$mgmt_prompt_regex = ZipTie::Adapters::Fortinet::FortiOS::AutoLogin::get_prompt( $cli_protocol );
	}

	my $mgmt_vdom_responses = _get_vdom_responses( $cli_protocol, $mgmt_prompt_regex );
	%responses = ( %responses, %{ $mgmt_vdom_responses } );

	# On VDOM enabled systems go to each non-mgmt VDOM and get the responses
	if( $vdoms_enabled ) {
		my $vdom_responses;
		my ($vdom_blob) = $responses{config} =~ /^config\s+vdom\s*(.+?)\s*end\s*$/ms;
		while( $vdom_blob =~ /^edit\s+(\S+)\s*$/mg ) {
			my $name = $1;
			next if $name eq $mgmt_vdom_name;

			$cli_protocol->send_and_wait_for( 'end', $prompt_regex );
			$cli_protocol->send_and_wait_for( 'config vdom', '\(vdom\)\s+#\s*$' );
			$cli_protocol->send_and_wait_for( "edit $name", "\\($name\\)\\s+#\\s*\$" );
			my $vdom_prompt_regex = ZipTie::Adapters::Fortinet::FortiOS::AutoLogin::get_prompt( $cli_protocol );
			$vdom_responses->{$name} = _get_vdom_responses( $cli_protocol, $vdom_prompt_regex );
		}
		%responses = ( %responses, ( 'vdoms', $vdom_responses ) );
	}

	if( $re_enable_paging ) {
		if( $vdoms_enabled ) {
			$cli_protocol->send_and_wait_for( 'end', $prompt_regex );
			_switch_to_global($cli_protocol);
		}
		_enable_paging( $cli_protocol, $global_prompt_regex );
	}

	clean_responses(\%responses);

	return \%responses;
}

sub _switch_to_global {
	my ($cli_protocol) = @_;

	my $commmand = "config global";
	
	my $response = $cli_protocol->send_and_wait_for( $commmand, '(#|\$)\s*$' );
	
	if ($response =~ /Unknown action|Command fail/i){
		$LOGGER->fatal("Unable to issue command: $commmand\n\nNB: A user with read/write access is needed to backup a device with virtual domains (VDOMs) enabled");
	}
		
	my $global_prompt_regex = ZipTie::Adapters::Fortinet::FortiOS::AutoLogin::get_prompt( $cli_protocol );

	return $global_prompt_regex;
}

sub _get_global_responses {
	my ($cli_protocol, $prompt_regex) = @_;
	my $responses;

	$responses->{system_interfaces} = $cli_protocol->send_and_wait_for( 'show system interface', $prompt_regex );
	$responses->{system_interfaces_dynamic} = $cli_protocol->send_and_wait_for( 'get system interface', $prompt_regex );
	$responses->{performance} = $cli_protocol->send_and_wait_for( 'get system performance status', $prompt_regex );
	$responses->{FQDNs} = $cli_protocol->send_and_wait_for( 'get test dnsproxy 6', $prompt_regex );

	# need to get responses for each nic here
	$responses->{hardware_nics} = $cli_protocol->send_and_wait_for( 'get hardware nic', $prompt_regex );
	my @hardwareNicsArray = get_hardware_nics($responses);
	foreach my $nic (@hardwareNicsArray){
		$responses->{nic_info}{$nic} = $cli_protocol->send_and_wait_for( "get hardware nic $nic", $prompt_regex );
	}

	return $responses;
}

sub _get_vdom_responses {
	my ($cli_protocol, $prompt_regex) = @_;
	my $responses;

	$responses->{staticRoutes} = $cli_protocol->send_and_wait_for( 'get router info routing-table all', $prompt_regex);
	$responses->{addresses} = $cli_protocol->send_and_wait_for( 'show firewall address', $prompt_regex );
	$responses->{addressGroups} = $cli_protocol->send_and_wait_for( 'show firewall addrgrp', $prompt_regex );
	
	# need to get responses for each predefined service here, removed in FortiOSv5
	$responses->{servicesPredefinedList} = $cli_protocol->send_and_wait_for( 'get firewall service predefined', $prompt_regex );
	my @servicesPrefinedList = get_services_predefined_list($responses);
	foreach my $service (@servicesPrefinedList){
		$responses->{servicesPredefined}{$service} = $cli_protocol->send_and_wait_for( "get firewall service predefined $service", $prompt_regex );
	}

	$responses->{servicesCustom} = $cli_protocol->send_and_wait_for( 'show firewall service custom', $prompt_regex );
	$responses->{serviceGroups} = $cli_protocol->send_and_wait_for( 'show firewall service group', $prompt_regex );
	$responses->{policies} = $cli_protocol->send_and_wait_for( 'show firewall policy', $prompt_regex );
	$responses->{zones} = $cli_protocol->send_and_wait_for( 'show system zone', $prompt_regex );
	$responses->{vips} = $cli_protocol->send_and_wait_for( 'show firewall vip', $prompt_regex );
	$responses->{vipGroups} = $cli_protocol->send_and_wait_for( 'show firewall vipgrp', $prompt_regex );
	$responses->{virtual_ip_pools} = $cli_protocol->send_and_wait_for( 'show firewall ippool', $prompt_regex );

	return $responses;
}

sub _disable_pagination {
	my ($cliProtocol, $promptRegex) = @_;
	
	my $re_enable = 0;
	my $consolePrompt = "\\(console\\)|$promptRegex";
	$cliProtocol->send_and_wait_for('config system console', $consolePrompt );
	# Don't want to configure the device permenantly, check what the output mode was before changing
	my $response = $cliProtocol->send_and_wait_for( 'get', $consolePrompt );
	$re_enable = $response =~ /^output\s+:\s+more\s*$/m;
	$cliProtocol->send_and_wait_for( 'set output standard', $consolePrompt );
	$cliProtocol->send_and_wait_for( 'end', $promptRegex );

	return $re_enable;
}

sub _enable_paging {
	my ($cliProtocol, $promptRegex) = @_;

	my $consolePrompt = "\\(console\\)|$promptRegex";
    $cliProtocol->send_and_wait_for('config system console', $consolePrompt );
	$cliProtocol->send_and_wait_for( 'set output more', $consolePrompt );
    $cliProtocol->send_and_wait_for( 'end', $promptRegex );

	return;
}

sub get_services_predefined_list {
	my $in = shift;

	my @servicesPredefined;
	foreach my $line ( split( /\n/, $in->{servicesPredefinedList} ) ) {
		$line =~ s/(^\s+|\s+$)//g;
		if ( $line =~ /^name:\s(\S+)/ ) {
			push @servicesPredefined, $1;
		}
	}
	return @servicesPredefined;
}

sub clean_responses {
	my $responses = shift;
	foreach my $command ( keys %$responses ) {
		unless ($command eq "policies" || $command eq "vips"){
			$responses->{$command} =~ s/^--More--(\s+)?//mg
		}else{
			# need to restore spacing for policy and vip parsing to work
			my $commandResponse = $responses->{$command};
			my $updateNextLine = 0;
			my $newCommandResponse;
			foreach my $line (split /\n/, $commandResponse ){
				if ($line =~ /^--More--/){
					$line =~ s/^--More--\s{10}//;
				}	
				$newCommandResponse .= "$line\n";
			}
			$responses->{$command} = $newCommandResponse; # update the response
		}
	}
	return;
}

1;
