package ZipTie::Adapters::PaloAlto::PANOS;

use strict;
use warnings;
use 5.010_000;

use ZipTie::ConnectionPath;
use ZipTie::Adapters::PaloAlto::PANOS::AutoLogin qw(get_connection get_response_objects get_response_object);
use ZipTie::Adapters::PaloAlto::PANOS::LookupTable;
use ZipTie::Adapters::PaloAlto::PANOS::Parsers qw(parse_system print_system create_config print_config parse_chassis print_chassis
  parse_interfaces print_interfaces parse_vsystems print_sub_elements parse_vlans print_vlans parse_virtual_routers
  parse_applications parse_predefined_and_shared_applications parse_shared_elements print_shared_object_groups);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

use Exporter qw(import);
our @EXPORT_OK = qw(get_user_group_list);

#use ZipTie::Adapters::PaloAlto::PANOS::GetApplications;

my $LOGGER = ZipTie::Logger::get_logger();

sub backup {
	my ($package_name, $backup_doc) = @_;

	$LOGGER->debug('Starting backup operation.');
	
	my ($connection_path) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );

	my $responses = _get_responses($connection_path);

	# Get QRadar version - (to determine if we model userFilters)
	$responses->{qradarVersion} = _get_qradar_version();

	# Instantiate the LookupTable object that will be used to pass information between parsers
	$responses->{lookup_tables} = ZipTie::Adapters::PaloAlto::PANOS::LookupTable->new();

	$LOGGER->debug('Opening Filehandle...');
	my $filehandle = get_model_filehandle( 'PANOS', $connection_path->get_ip_address() );

	$LOGGER->debug('Opening Printer...');
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	$LOGGER->debug('Parsing System Information...');
	parse_system( $responses, $printer );
	print_system( $responses, $printer );

	$LOGGER->debug('Parsing Chassis Information...');
	parse_chassis( $responses, $printer );
	print_chassis( $responses, $printer );

	$LOGGER->debug('Parsing Configuration Information...');
	create_config( $responses, $printer );
	print_config( $responses, $printer );

	$LOGGER->debug('Parsing VLAN Information...');
	parse_vlans( $responses, $printer );
	# Vlans are printed later

	# The "admin virtual system" shared area definitions need to be parsed somewhere here.  It shows
	# up variously elsewhere in the parsing, but we're focusing here on things which will
	# need reference/resolution later in the parsing process.  The parsing here may be moved up in 
	# the event other definitions here need to be referenced earlier in the "print" process.  This 
	# is simply something which needs to go into 'objectGroups' and therefore needs to be printed
	# at this point as of this writing.
	$LOGGER->debug('Parsing shared "global" scope information...');
	parse_shared_elements($responses, $printer);
	print_shared_object_groups($responses, $printer);

	$LOGGER->debug('Parsing Interface Information...');
	parse_interfaces( $responses, $printer );
	print_interfaces( $responses, $printer );

	$LOGGER->debug('Parsing Predefined and Shared Application Information...');
	parse_predefined_and_shared_applications($responses);

	$LOGGER->debug('Parsing Sub-Element Information...');
	parse_virtual_routers($responses);
	parse_vsystems($responses);
	print_sub_elements( $responses, $printer );

	# Parse vlan information before interface information so that Vlan associatings are available to include
	# in interfaceVlanTrunks elements
	print_vlans( $responses, $printer );

	$LOGGER->debug('Closing Printer...');
	$printer->close_model;
	$LOGGER->debug('Printer closed');

	close_model_filehandle($filehandle);
	$LOGGER->debug('Filehandle closed');

	$LOGGER->debug('Backup operation complete.');

	return;
}

sub _get_responses {
	my ($connection_path) = @_;

	my $connection = get_connection($connection_path);

	my %commands = (
		required => [
			[ show_system_info    => 'type=op&cmd=<show><system><info></info></system></show>' ],
			[ show_config_running => 'type=op&cmd=<show><config><running></running></config></show>' ],
			[ show_interface_all  => 'type=op&cmd=<show><interface>all</interface></show>' ]
		],
		optional => [
			[ show_system_resources          => 'type=op&cmd=<show><system><resources></resources></system></show>' ],
			[ show_dynamic_address_groups    => 'type=op&cmd=<show><object><dynamic-address-group><all></all></dynamic-address-group></object></show>' ],
			[ show_predefined_applications   => 'type=config&action=get&xpath=/config/predefined/application' ],
			[ show_predefined_service_groups => 'type=config&action=get&xpath=/config/predefined/service' ],
			[ show_config_panorama           => 'type=config&action=get&xpath=/config/panorama' ]
		]
	);

	my $responses = get_response_objects($connection, \%commands);

	if( $connection_path->get_management_server_ip_address() ) {
		$responses->{is_proxy_backup} = 1;
	}

	# Dynamic command, we do not know a parameter until parsing starts
	$responses->{request_dynamic_block_list} = sub {
		my ($name, $type, $version) = @_;
		my $response;
		
		# If $type was passed in then this is v7.1+ we need to send a slightly different request
		if( $version gt '7.1' ) {
			$response = get_response_object(
				$connection,
				"request system external-list show type $type name $name",
				"type=op&cmd=<request><system><external-list><show><type><$type><name>$name</name></$type></type></show></external-list></system></request>"
			);
		} else {
			$response = get_response_object(
				$connection,
				"request system external-list show name $name",
				"type=op&cmd=<request><system><external-list><show><name>$name</name></show></external-list></system></request>"
			);
		}

		return $response;
	};

	$responses->{request_fqdn} = sub {
		my ($vsys, $name) = @_;
		state $sleeps = 0;

		REQUEST_FQDN_EVAL:
		my $response = get_response_object(
			$connection,
			"request system fqdn show-object $name",
			"type=op&cmd=<request><system><fqdn><show-object><vsys>$vsys</vsys><name>$name</name></show-object></fqdn></system></request>"
		);

		# PAN-OS will not let you request FQDNs while it is refreshing them
		# sleep then try again a few times
		if( $response ) {
			my $msg = $response->findvalue('/response/result/msg/line');
			if( $msg and $msg =~ /refresh/i ) {
				if( $sleeps < 3 ) {
					sleep 20; ++$sleeps; goto REQUEST_FQDN_EVAL;
				}
			}
		}
		return $response;
	};

	return $responses;
}

=begin to be used in future user group modelling
sub get_user_group_list {
	my $responses = shift;

	my $show_user_group_list = $responses->{show_user_group_list};
	my $userGroupListResponse = $show_user_group_list->findvalue('/response/result');
	my @userGroups = split (/\s(?=cn)/, $userGroupListResponse);
	my @cleanUserGroups = grep { $_ ne ''} @userGroups; # remove any empty values
	$cleanUserGroups[-1] =~ s/\s+Total:\s\d+\s+//; # remove total from last value
	for my $userGroup (@cleanUserGroups) { # remove unwanted whitespace
		$userGroup =~ s/^\s+|\s+$//;
	}
	return @cleanUserGroups;
}
=cut

sub _get_qradar_version {
	my $version = `/opt/qradar/bin/myver`;
	$version =~ s/\.//g;
	$version =~ /(\d\d\d)/;
	return $1;
}

1;
