package ZipTie::Adapters::PaloAlto::PANOS::AutoLogin;

use strict;
use warnings;
use 5.010_000;

use XML::Twig::XPath;

use ZipTie::Adapters::PaloAlto::PANOS::Utils qw(get_value_from_xml);
use ZipTie::Adapters::Utils qw(parseCIDR get_crep);
use ZipTie::Logger;
use ZipTie::RESTAPI;

our @EXPORT_OK = qw(get_connection get_response_objects get_response_object); 
use Exporter qw(import);

my $LOGGER = ZipTie::Logger::get_logger();

my $TOKEN_QUERY;

sub get_connection {
	my ($connection_path) = @_;

	my $connection = ZipTie::RESTAPI->new($connection_path, '/api/?');

        my $credentials = $connection_path->get_credentials();
        my $username = $credentials->{username};
        my $password = $credentials->{password};
        $LOGGER->debug("Using username: $username");

	$TOKEN_QUERY = "type=keygen&user=$username&password=$password";
        $connection->obtain_token_as_parameter($TOKEN_QUERY, '', \&_parse_connection_key);

	# Is it a proxied backup
	if( $connection_path->get_management_server_ip_address() ) {
		my $target_ip = $connection_path->get_ip_address();
		my $mgmt_ip = $connection_path->get_management_server_ip_address();
		$connection->obtain_target_as_parameter('type=op&cmd=<show><devices><connected></connected></devices></show>', \&_parse_target, $target_ip, $mgmt_ip);
	}

	return $connection;
}

sub get_response_objects {
	my ($connection, $commands) = @_;
	my $responses;

	my @required_commands = @{ $commands->{required} };
	for my $command ( @required_commands ) {
		my ($name, $endpoint) = @{ $command };
		my $response = get_response_object( $connection, $name, $endpoint );

		if( not $response ) {
			$LOGGER->fatal("\nERROR: The command '$name' failed. It is required for successful backup.");
		}

		# Check this isn't a Panorama that is being backed up, people sometimes do this
		# It is the first command sent after authentication to fail early
		if( $name eq 'show_system_info' ) {
			my $model = $response->findvalue('/response/result/system/model');
			if( $model and $model =~ /Panorama/i ) {
				$LOGGER->fatal("\nERROR: Cannot perform this opertion on a Panorama management server.\nYou must discover managed firewalls from the Panorama sever and then backup those devices.");
			}
		}

		$responses->{$name} = $response;
	}

	# Optional commands are themselves optional, e.g. Applications does not have any optional commands
	if( $commands->{optional} ) {
		my @optional_commands = @{ $commands->{optional} };
		for my $command ( @optional_commands ) {
			my ($name, $endpoint) = @{ $command };

			my $response = $responses->{$name} = get_response_object( $connection, $name, $endpoint);

			if( not $response ) {
				$LOGGER->debug("Retrieval of '$name' failed. This is not a required command so the operation can proceed.");
			}
		}
	}

	return $responses;
}

sub get_response_object {
	my ($connection, $command_name, $command_endpoint) = @_;
	
	my $response =  $connection->get($command_name, $command_endpoint);

	# Session can timeout, this renegotiates and starts this iteration again
	if( $response =~ />Session timed out</i ) {
		$LOGGER->debug('Session has timed out, requesting a new session token');
		$connection->obtain_token_as_parameter($TOKEN_QUERY, '', \&_parse_connection_key);
		$LOGGER->debug("Resending command '$command_endpoint'");
		redo;
	}
	
	return _validate_response($response, $command_name);
}

sub _parse_connection_key {
        my ($xml) = @_;

        if($xml !~ /</){
                $LOGGER->fatal("Could not reach device: $xml\nCheck your credentials and protocol settings.\n");
        }

        my $twig = XML::Twig::XPath->new();
        $twig->parse($xml);

        my $attributes = $twig->root()->atts();
        if( $attributes->{status} ne 'success' ) {
                my $msg = $twig->findvalue('/response/result/msg');
                $LOGGER->fatal("REST API CALL FAILED: $msg\n");
        }

        if( $twig->findnodes('/response/result/error') ) {
                my $msg = $twig->findvalue('/response/result/error');
                $LOGGER->fatal("REST API CALL FAILED: $msg\n");
        }

        return '&key=' . $twig->findvalue('/response/result/key');
}

sub _parse_target {
	my ($xml, $target_ip, $mgmt_ip) = @_;
	my $target;

	my $twig = _validate_response($xml, 'show_devices_connected');

	my $serial = get_value_from_xml( "/response/result/devices/entry[ip-address='$target_ip']/serial", $twig );
	if( defined $serial ) {
		$target = "&target=$serial";
	} else {
		$LOGGER->fatal( "Device with IP '$target_ip' is not currently connected to the management server at '$mgmt_ip'" );
	}

	return $target;
}

sub _validate_response {
	my ($xml, $command) = @_;

	my $eval_success;
	my $twig;
	$eval_success = eval {
		$twig = XML::Twig::XPath->new()->parse($xml);
		1; # Magic true to set eval_success
	};
	if( not $eval_success ) {
		$LOGGER->debug("REST API CALL '$command' FAILED due to XML parsing error: $@");
	} else {
		my $attributes = $twig->root()->atts();
		if( $attributes->{status} ne 'success' ) {
			my $msg = $twig->findvalue('/response/result/msg/line');
			$LOGGER->debug("REST API CALL '$command' FAILED: $msg");
		} elsif( $twig->findnodes('/response/result/error') ) {
			my $error = $twig->findvalue('/response/result/error');
			$LOGGER->debug("REST API CALL '$command' FAILED: $error");
		}
	}
	return $twig;
}

1;
