package ZipTie::Adapters::PaloAlto::PANOS::LookupTable;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
$Data::Dumper::Terse = 1; # Do not print $VAR1 before each dumped variable
$Data::Dumper::Sortkeys = 1; # Sort hash keys in dumped hashes

use ZipTie::Logger;
my $LOGGER = ZipTie::Logger::get_logger();

sub new {
	my ($class) = @_;

	my $self = {
		interfaces_by_name           => {},
		interfaces_by_vsys           => {},
		zones_by_vsys                => {},
		vlans_by_interface           => {},
		virtual_router_by_name       => {},
		network_groups               => {},
		service_groups               => {},
		applications                 => {},
		application_filters          => {},
		application_groups           => {},
		default_port_service_groups  => {},
		ipsec_tunnels                => {},
		management_profiles          => {},
		os_version                   => "",
		global_mtu                   => 1500 # default mtu
	};

	bless $self, $class;
	return $self;
}

sub add_interface {
	my ( $self, $name, $interface ) = @_;
	$self->{interfaces_by_name}->{$name} = $interface;
	return;
}

sub get_interface {
	my ( $self, $name ) = @_;
	return $self->{interfaces_by_name}->{$name};
}

sub add_interface_to_vsys { 
	my ( $self, $vsys_id, $interface_name, $is_shared_gateway ) = @_;
	my $vsys_name = $is_shared_gateway ? "sg$vsys_id" : "vsys$vsys_id";
	$self->{interfaces_by_vsys}->{$vsys_name}->{$interface_name} = 1;
	return;
}

sub get_interfaces_by_vsys {
	my ( $self, $vsys_name ) = @_;
	my @interfaces = keys %{ $self->{interfaces_by_vsys}->{$vsys_name} };
	return \@interfaces;
}

sub add_zone_to_vsys {
	my ( $self, $zone_name, $vsys_name, $interface_list ) = @_;
	$self->{zones_by_vsys}->{$vsys_name}->{$zone_name} = $interface_list;
	return;
}

sub get_zones_by_vsys {
	my ( $self, $vsys_name ) = @_;
	return $self->{zones_by_vsys}->{$vsys_name};
}

sub add_virtual_router {
	my ( $self, $name, $virtual_router ) = @_;
	$self->{virtual_router_by_name}->{$name} = $virtual_router;
	return;
}

sub get_virtual_router {
	my ( $self, $name ) = @_;

	return $self->{virtual_router_by_name}->{$name};
}

sub add_vlan_member {
	my ( $self, $name, $interface ) = @_;

	push @{$self->{'vlans_by_interface'}->{$interface}}, $name;
	return;
}

sub get_interface_vlan_membership {
	my ( $self, $interface ) = @_;

	if (defined $self->{'vlans_by_interface'}->{$interface}) {
		return $self->{'vlans_by_interface'}->{$interface};
	}
	else {
		return;
	}
}

sub add_virtual_wire {
	my ( $self, $name, $first_interface, $second_interface ) = @_;

	$self->{'virtual_wires'}->{$first_interface} = $name;
	$self->{'virtual_wires'}->{$second_interface} = $name;

	return;
}

sub get_virtual_wire {
	my ( $self, $interface_name ) = @_;
	
	if (defined $self->{'virtual_wires'}->{$interface_name}) {
		return $self->{'virtual_wires'}->{$interface_name};
	}
	else {
		return;
	}
}


sub add_service_group {
	my ( $self, $name, $gObject ) = @_;

	my $base_name;
	#determine if the service group to be added is a source half or a destination half.
	#add the group to a hash indexed by the base name of the service group so that the two
	#halves can be related to each other easily
	if ( $name =~ /_source$/ ) {
		($base_name) = $name =~ /(.*)_source/;
		$self->{service_groups}->{$base_name}->{source} = $name;
	} elsif ( $name =~ /_destination$/ ) {
		($base_name) = $name =~ /(.*)_destination/;
		$self->{service_groups}->{$base_name}->{destination} = $name;
	} else {
		$LOGGER->debug("WARNING: Parsing of service group '$name' failed. Source/destination components not setup correctly.");
	}
	
	$self->{service_groups}->{$base_name}->{object_exp} = $gObject if ($gObject);
	
	return;
}

sub get_service_group {
	my ( $self, $base_name ) = @_;
	if ( !defined $self->{service_groups}->{$base_name} ) {
		$LOGGER->debug("WARNING: Service group '$base_name' not found in the lookup table.");
		return;
	}
	return $self->{service_groups}->{$base_name};
}

sub get_service_group_object {
	my ( $self, $base_name ) = @_;
	if ( !defined $self->{service_groups}->{$base_name} ) {
		# try to strip 'source' or 'destination' off of the base name
		if ($base_name =~ /^(.+?)_(source|destination)/)
		{
			$base_name = $1;
		}
		else
		{
			$LOGGER->debug("WARNING: Service group '$base_name' not found in the lookup table.");
			return {};
		}
	}
	# return either object group expansion or a degenerate shell.
	return $self->{service_groups}->{$base_name}->{object_exp} // { id => $self->{service_groups}->{$base_name} };
}

sub get_service_group_direction_set {
	my ($self, $base_name) = @_;	
	my @sgSet;
	
	if ($self->get_service_group($base_name))
	{
		for (values (%{$self->{service_groups}->{$base_name}}))
		{
			push @sgSet, $_;
		}
	}
	return \@sgSet;
}

sub add_network_group {
	my ($self, $name, $group) = @_;
	$self->{network_groups}->{$name} = $group;
	return;
}

sub get_network_group {
	my ($self, $name) = @_;
	return $self->{network_groups}->{$name};
}

sub add_application{
	my ($self, $name, $application) = @_;

	if(!$self->{applications}->{$name}){
		$self->{applications}->{$name} = $application;
	}else{
		$LOGGER->debug("WARNING: Application '$name' already exists, not overwriting");
	}
	return;
}

sub get_applications{
	my ($self) = @_;
	return($self->{applications});
}

sub add_application_group {
	my ($self, $name, $group) = @_;

	if(!$self->{application_groups}->{$name}){
		$self->{application_groups}->{$name} = $group;
	}else{
		$LOGGER->debug("WARNING: Application group '$name' already exists, not overwriting");
	}
	return;
}

sub is_application_group_defined {
	my ($self, $name) = @_;

	if (defined $self->{application_groups}->{$name}) {
		return 1;
	}
	return;
}

sub add_application_filter{
	my ($self, $name, $filter) = @_;

	if(!$self->{application_filters}->{$name}){
		$self->{application_filters}->{$name} = $filter;
	}else{
		$LOGGER->debug("WARNING: Application filter '$name' already exists, not overwriting");
	}
	return;
}

sub get_application_filter{
	my ($self, $name) = @_;

	if($self->{application_filters}->{$name}){
		return($self->{application_filters}->{$name});
	}
	return;
}

=begin to be used for future user group modelling
sub add_user_group {
	my ($self, $name, $group, $shortName) = @_;

	if(!$self->{user_groups}->{$name}){
		$self->{user_groups}->{$name} = $group;
		$self->{user_groups}->{$name}->{shortName} = $shortName if defined $shortName;
	}else{
		$LOGGER->debug("WARNING: User group '$name' already exists, not overwriting");
	}
	return;
}

sub is_user_group_defined {
	my ($self, $name) = @_;

	if (defined $self->{user_groups}->{$name}) {
		return 1;
	}
	return;
}
sub get_user_shortname {
	my ($self, $name) = @_;
	if (defined $self->{user_groups}->{$name}->{shortName}) {
		return $self->{user_groups}->{$name}->{shortName};
	} else {
		return "";
	}
}
=cut

sub add_default_port_service_group{
	my ($self, $name) = @_;

	# Do not attempt to create a new service group if we have already created it
	# Unless we do not have any defined, then don't look for the exisitng service group
	if( $self->{default_port_service_groups} and $self->{default_port_service_groups}->{$name} ) {
		$LOGGER->verbose("Default service already created for application '$name'");
	} else {
		my $new_service_group = {
			id => $name,
			description => "The default service(s) for $name"
		};

		if($self->{applications}->{$name}){
			my $application = $self->{applications}->{$name};
			my @keys = keys(%{$application->{service_group}});
			foreach my $key(@keys){
				foreach my $service_group_value(@{$application->{service_group}->{$key}}){
					push(@{$new_service_group->{$key}}, $service_group_value);
				}
			}

		}elsif($self->{application_groups}->{$name}){
			my $application_group = $self->{application_groups}->{$name};
			$self->_dedupe_application_services($new_service_group, $application_group);

			foreach my $group_name(@{$application_group->{objectGroupReference}}){
				$self->add_default_port_service_group($group_name);
				push(@{$new_service_group->{objectGroupReference}}, $group_name);
			}

		}elsif($self->{application_filters}->{$name}){
			my $application_filter = $self->{application_filters}->{$name};
			$self->_dedupe_application_services($new_service_group, $application_filter);
		}

		$self->{default_port_service_groups}->{$name} = $new_service_group;
		$LOGGER->verbose("Created default service for application '$name'");
		$LOGGER->verbose('service = ' . Dumper($new_service_group));
	}

	return;
}

sub _dedupe_application_services {
	my ($self, $new_service_group, $applications) = @_;

	foreach my $application_name(@{$applications->{application}}){
		my $application = $self->{applications}->{$application_name};

		for my $service_key ( keys %{ $application->{service_group} } ) {
			for my $service ( @{ $application->{service_group}->{$service_key} } ) {
				if( not $new_service_group->{$service_key} ) {
					$new_service_group->{$service_key} = [ $service ];
				} else {
					my $already_exists = 0;
					for my $existing_service ( @{ $new_service_group->{$service_key} } ) {
						my $match = 1;
						for my $key ( keys %{ $service } ) {
							if( $service->{$key} ne $existing_service->{$key} ) {
								$match = 0; last;
							}
						}
						if( $match ) {
							$already_exists = 1; last;
						}
					}
					if( not $already_exists ) {
						push @{ $new_service_group->{$service_key} }, $service;
					}
				}
			}
		}
	}
	return $new_service_group;
}

sub get_default_port_service_groups{
	my ($self) = @_;
	my @service_groups;
	
	foreach my $key(keys(%{$self->{default_port_service_groups}})){
		push(@service_groups, $self->{default_port_service_groups}->{$key});
	}

	return(\@service_groups);
}

sub add_ipsec_tunnel {
	my ($self, $name, $entry) = @_;
	$self->{ipsec_tunnels}->{$name} = $entry;
	return;
}

sub get_ipsec_tunnel {
	my ($self, $name) = @_;
	my $ipsec_tunnel;
	if( $self->{ipsec_tunnels}->{$name} ) {
		$ipsec_tunnel = $self->{ipsec_tunnels}->{$name};
	}
	return $ipsec_tunnel;
}

sub add_interface_management_profile{
	my ($self, $name, $network_group, $service_group) = @_;

	$self->{management_profiles}->{$name} = {
		network_group => $network_group,
		service_group => $service_group
	};
}

sub add_interface_to_management_profile{
	my ($self, $profile_name, $interface_name) = @_;

	if(!defined $self->{management_profiles}->{$profile_name}){
		$LOGGER->debug("WARNING: Interface Management Profile '$profile_name' not found in the lookup table, cannot add '$interface_name'");
	}else{
		push(@{$self->{management_profiles}->{$profile_name}->{interfaces}}, $interface_name);
	}
}

sub get_interface_management_profiles{
	my ($self) = @_;

	my @profiles = ();
	foreach my $profile_key(keys(%{$self->{management_profiles}})){
		$self->{management_profiles}->{$profile_key}->{name} = $profile_key;
		push(@profiles, $self->{management_profiles}->{$profile_key});
	}
	return(\@profiles);
}

sub set_os_version {
	my ($self, $version) = @_;
	
	if ($version && $version =~ /(\d.\d)/){
		$self->{os_version} = $1;
	}
	return;
}

sub get_os_version{
	my ($self) = @_;

	return($self->{os_version});
}


sub set_global_mtu {
	my ($self, $global_mtu) = @_;
	$self->{global_mtu} = $global_mtu;
	return;
}

sub get_global_mtu {
	my ($self) = @_;
	return $self->{global_mtu};
}

1;
