package ZipTie::Adapters::PaloAlto::PANOS::Parsers;

use strict;
use warnings;
use 5.010_000;

use Net::CIDR::Set;
use MIME::Base64 qw(encode_base64);

use ZipTie::Adapters::PaloAlto::PANOS::Rulebase qw(parse_acls print_filter_lists parse_nat);
use ZipTie::Adapters::PaloAlto::PANOS::TrimServiceGroups qw(trim_service_groups);
use ZipTie::Adapters::PaloAlto::PANOS::Utils qw(get_value_from_xml print_if_defined unwind_port_expression parse_cidr_panos get_interface_type_panos add_interface_grouping);
use ZipTie::Adapters::Utils qw(strip_mac trim mask_to_bits create_network_element invert_mask get_mask get_port_expression get_crep);
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_system print_system create_config print_config parse_chassis print_chassis
  parse_interfaces print_interfaces parse_vsystems print_sub_elements parse_vlans print_vlans parse_virtual_routers
  get_serial_number parse_applications parse_predefined_and_shared_applications parse_shared_elements print_shared_object_groups);

my $LOGGER = ZipTie::Logger::get_logger();

my $IP_REGEX = get_crep('cipm');

sub parse_system {
	my ( $in, $out ) = @_;

	# Grab the XML Twigs that we need
	my $show_system_info	= $in->{show_system_info};
	my $show_config_running = $in->{show_config_running};

	# Find the system name
	my $system_name = get_value_from_xml( '/response/result/system/devicename', $show_system_info );
	$in->{parsed}->{'core:systemName'} = $system_name;

	$in->{parsed}->{'core:deviceType'} = 'Firewall';
	$in->{parsed}->{'core:osInfo'}->{'core:name'} = 'PAN-OS';

	#Find the software version
	my $os_version = get_value_from_xml( '/response/result/system/sw-version', $show_system_info );
	$in->{parsed}->{'core:osInfo'}->{'core:version'} = $in->{version} = $os_version;
	$in->{lookup_tables}->set_os_version($os_version); # store for later lookup

	return;
}

sub print_system {
	my ( $in, $out ) = @_;

	print_if_defined( $in->{parsed}, 'core:systemName', $out );

	$out->open_element('core:osInfo');

	$out->print_element( 'core:make', 'Palo Alto' );

	print_if_defined( $in->{parsed}->{'core:osInfo'}, 'core:name',          $out );
	print_if_defined( $in->{parsed}->{'core:osInfo'}, 'core:softwareImage', $out );
	print_if_defined( $in->{parsed}->{'core:osInfo'}, 'core:version',       $out );

	$out->print_element( 'core:osType', 'PAN-OS' );

	_print_packet_process_ordering($out);

	$out->close_element('core:osInfo');

	print_if_defined( $in->{parsed}, 'core:biosVersion', $out );
	print_if_defined( $in->{parsed}, 'core:deviceType',  $out );
	print_if_defined( $in->{parsed}, 'core:contact',     $out );
	print_if_defined( $in->{parsed}, 'core:lastReboot',  $out );
	return;
}

sub parse_chassis {
	my ( $in, $out ) = @_;

	#Grab the XML Twigs that we need
	my $show_system_info	= $in->{show_system_info};
	my $show_config_running = $in->{show_config_running};

	#The asset we are creating is a Chassis so this can be hard coded
	$in->{parsed}->{chassis}->{'core:asset'}->{'core:assetType'} = 'Chassis';

	#This adapter will only handle Palo Alto devices
	$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = 'Palo Alto';

	#Find the model number
	my $model_number = get_value_from_xml( '/response/result/system/model', $show_system_info );
	if( $model_number ) {
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $model_number;
	}

	#Find the Serial Number
	#Do not allow 'unknown', will not allow more than one PANOS VM to appear in topology
	my $serial_number = get_value_from_xml( '/response/result/system/serial', $show_system_info );
	if( $serial_number ) {
		if( $serial_number !~ /^unknown$/i ) {
			$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $serial_number;
		} else {
			$LOGGER->warning($INVALID_CONFIGURATION, "Serial number '$serial_number' will not be modelled because it can be identical for different virtual PANOS instances.");
		}
	}

	#Find the device's location
	my $location = get_value_from_xml( '/response/result/config/devices/entry/deviceconfig/system/snmp-setting/snmp-system/location', $show_config_running );
	if( $location ) {
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:location'}->{'core:description'} = $location;
	}

	#Find the System level mac address
	my $mac_address = get_value_from_xml( '/response/result/system/mac-address', $show_system_info );
	if( $mac_address ) {
		$in->{parsed}->{chassis}->{'macAddress'} = strip_mac($mac_address);
	}

	#Find the system memory
	if( defined $in->{show_system_resources} ) {
		my $system_resources = get_value_from_xml( '/response/result', $in->{show_system_resources} );
		if( defined $system_resources ) {
			my ($size) = $system_resources =~ /Mem:\s*(\d+)k\s*total/i;

			$size *= 1024;

			$in->{parsed}->{chassis}->{memory}->{kind} = 'RAM';
			$in->{parsed}->{chassis}->{memory}->{size} = $size;
		}
	}

	return;
}

sub print_chassis {
	my ( $in, $out ) = @_;
	$out->open_element('chassis');
	print_if_defined( $in->{parsed}->{chassis}, 'core:asset',  $out );
	print_if_defined( $in->{parsed}->{chassis}, 'macAddress',  $out );
	print_if_defined( $in->{parsed}->{chassis}, 'memory',	   $out );
	print_if_defined( $in->{parsed}->{chassis}, 'powersupply', $out );
	$out->close_element('chassis');
	return;
}

sub create_config {
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository;
	$repository->{'core:name'} = '/';

	# build the simple text configuration
	$in->{'show_config_running'}->set_pretty_print('indented');

	{
		my $running;
		$running->{'core:name'} = 'running-config';

		#we encode to utf8, because we have no assurance that there are not crazy characters, encode_base64 will barf with the following "Wide character in subroutine entry"
		$running->{'core:textBlob'}   = encode_base64( Encode::encode_utf8( $in->{'show_config_running'}->first_elt('config')->sprint() ));
		$running->{'core:mediaType'}  = 'text/xml';
		$running->{'core:context'}    = 'active';
		$running->{'core:promotable'} = 'false';

		# push the configuration into the repository
		my $system_folder;
		$system_folder->{'name'} = 'system space';
		push @{$system_folder->{'core:config'}}, $running;
		push @{$repository->{'core:folder'}}, $system_folder;

		#also push it at the base level
		my $test_folder;
		$test_folder->{'name'} = 'admin';
		push @{$test_folder->{'core:config'}}, $running;
		push @{$repository->{'core:folder'}}, $test_folder;
	}

	#Extract vsys entries from the running config
	my @virtual_systems = $in->{'show_config_running'}->findnodes('/response/result/config/devices/entry/vsys/entry');
	foreach my $vsys (@virtual_systems) {
		my $running;
		my $folder;
		$folder->{name} = $vsys->att('name');

		$running->{'core:name'}		  = 'running-config';
		#we encode to utf8, because we have no assurance that there are not crazy characters, encode_base64 will barf with the following "Wide character in subroutine entry"
		$running->{'core:textBlob'}   =  encode_base64( Encode::encode_utf8( $vsys->sprint() ));
		$running->{'core:mediaType'}  = 'text/xml';
		$running->{'core:context'}	  = 'active';
		$running->{'core:promotable'} = 'false';

		# push the configuration into the repository
		push @{ $folder->{'core:config'} }, $running;
		push @{ $repository->{'core:folder'} }, $folder;
	}

#	#Extract virtual router entries from the running config
#	my @virtual_routers = $in->{'show_config_running'}->findnodes('/response/result/config/devices/entry/network/virtual-router/entry');
#	foreach my $router (@virtual_routers) {
#		my $running;
#		my $folder;
#		$folder->{name} = $router->att('name');
#
#		$running->{'core:name'}		  = 'running-config';
#		$running->{'core:textBlob'}   = encode_base64( $router->sprint() );
#		$running->{'core:mediaType'}  = 'text/xml';
#		$running->{'core:context'}	  = 'active';
#		$running->{'core:promotable'} = 'false';
#
#		# push the configuration into the repository
#		push @{ $folder->{'core:config'} }, $running;
#		push @{ $repository->{'core:folder'} }, $folder;
#	}

	# save the repository
	$in->{parsed}->{'core:configRepository'} = $repository;
	return;
}

sub print_config {
	my ( $in, $out ) = @_;

	$out->open_element('core:configRepository name="'.$in->{parsed}->{'core:configRepository'}->{'core:name'}.'"');

	#print the configs at the base level, ie the complete running config
	foreach my $config (@{$in->{parsed}->{'core:configRepository'}->{'core:config'}}) {
		$out->print_element('core:config', $config);
	}

	#print the sub-configs, ie the configs for virtual systems and virtual routers
	foreach my $folder (@{$in->{parsed}->{'core:configRepository'}->{'core:folder'}}) {
		$out->open_element('core:folder name="'.$folder->{name}.'"');
		foreach my $config (@{$folder->{'core:config'}}) {
			$out->print_element('core:config', $config);
		}
		$out->close_element('core:folder');
	}

	$out->close_element('core:configRepository');

	return;
}

sub parse_interfaces {
	my ( $in, $out ) = @_;

	my $show_interface_all  = $in->{show_interface_all};
	my $show_config_running = $in->{show_config_running};

	## Parse data into the lookup object
	# Set up virtual wire information in the lookup table, used in _parse_interface
	my @virtual_wires = $show_config_running->find_nodes('/response/result/config/devices/entry/network/virtual-wire/entry');
	_parse_virtual_wires(\@virtual_wires, $in->{lookup_tables});

	# Create interface management profile information in the lookup table, used in object group parsing
	my @management_profiles = $show_config_running->find_nodes('/response/result/config/devices/entry/network/profiles/interface-management-profile/entry');
	_parse_management_profile(\@management_profiles, $in->{lookup_tables});

	# Get the global mtu value
	my $global_mtu = get_value_from_xml('/response/result/config/devices/entry/deviceconfig/setting/jumbo-frame/mtu', $show_config_running);
	if( defined $global_mtu ) {
		$in->{lookup_tables}->set_global_mtu($global_mtu);
	}

	# Parse the configured ipsec tunnels so the information is available when parsing tunnel interfaces
	my @tunnel_entrys = $show_config_running->findnodes('/response/result/config/devices/entry/network/tunnel/ipsec/entry');
	my @gateway_entrys = $show_config_running->findnodes('/response/result/config/devices/entry/network/ike/gateway/entry');
	_parse_ipsec_tunnels(\@tunnel_entrys, \@gateway_entrys, $in->{lookup_tables});
	##


	## Start actual interface parsing
	# Loop through interface (and their subinterfaces) deined in 'show config running' adding each to the lookup object
	my @interface_types = qw(ethernet/entry vlan loopback tunnel); # aggregate-ethernet is another type but is not present here in 6.0, so use old parsing
	for my $interface_type ( @interface_types ) {
		my @interface_configs = $show_config_running->find_nodes("/response/result/config/devices/entry/network/interface/$interface_type");

		$interface_type =~ s/\/.+//; # remove '/entry' for ethernet

		for my $interface_config (@interface_configs) {
			_parse_interface( $interface_config, $interface_type, 'primaryInterface', $in->{lookup_tables}, $show_interface_all );
			
			my @sub_interfaces_config = $interface_config->findnodes('.//units/entry');
			for my $sub_interface_config ( @sub_interfaces_config ) {
				_parse_interface( $sub_interface_config, $interface_type, 'subInterface', $in->{lookup_tables}, $show_interface_all );
			}
		}
	}

	my @ifnet_interfaces = $show_interface_all->find_nodes('/response/result/ifnet/entry');
	my @primary_interfaces;
	for my $ifnet_interface ( @ifnet_interfaces ) {

		my $name = get_value_from_xml('name', $ifnet_interface);
		my $interface = $in->{lookup_tables}->get_interface($name);

		if( not $interface ){
			$LOGGER->verbose("Interface '$name' does not exist, creating a new interface");

			$interface = { name => $name };

			$interface->{interfaceType} = get_interface_type_panos($name);
			$interface->{physical} = $interface->{interfaceType} eq 'ethernet' ? 'true' : 'false';

			add_interface_grouping( $interface, 'any', 'securityZoneMember' );

			my $vsys_id = get_value_from_xml('vsys', $ifnet_interface);
			if( defined $vsys_id ) {
				$in->{lookup_tables}->add_interface_to_vsys( $vsys_id, $name );
			}

			$in->{lookup_tables}->add_interface( $name, $interface );
		}

		if($name =~ /^(.*?)\.\d+$/){
			my $parent_name = $1;
			add_interface_grouping( $interface, $parent_name, 'subInterface' );

			# Remember to add the primary interface later
			push(@primary_interfaces, $parent_name);
		}

		my $zone = get_value_from_xml('zone', $ifnet_interface);
		if($zone){
			add_interface_grouping($interface, $zone, 'securityZoneMember');
		}

		# Add any IP addresses we did not previously add
		# These include address objects, that are resolved by PANOS in "show interface all"
		my $ip = get_value_from_xml('ip', $ifnet_interface);
		my @addresses = ($ip);
		my @addr_members = $ifnet_interface->findnodes('addr/member');
		for my $addr_member ( @addr_members ) {
			push @addresses, get_value_from_xml('.', $addr_member);
		}

		for my $address (@addresses) {
			if($address =~ /^((\d+\.)+\d+)\/(\d+)$/){
				my $ip_address = $1;
				my $mask = $3 // 32;

				my $ip_configuration = { ipAddress => $ip_address };
				if($mask != 32){
					$ip_configuration->{mask} = $mask;
				}

				my $add = 1;
				foreach my $ip_config(@{$interface->{interfaceIp}->{ipConfiguration}}){
					if($ip_config->{ipAddress} eq $ip_address){
						$add = 0; last;
					}
				}
				if($add){
					push(@{$interface->{interfaceIp}->{ipConfiguration}}, $ip_configuration);
				}
			}
		}
	}

	my @hardware_interfaces = $show_interface_all->find_nodes('/response/result/hw/entry');
	my $aggregate_interfaces;

	for my $hardware_interface ( @hardware_interfaces ) {
		my $name = get_value_from_xml('name', $hardware_interface);
		my $interface = $in->{lookup_tables}->get_interface($name);

		if( not $interface ) {
			$LOGGER->verbose("Interface '$name' does not exist, creating a new interface");

			$interface = { 
				name => $name
			};
			
			# Infer interface type and physicality
			$interface->{interfaceType} = get_interface_type_panos($name);
			$interface->{physical} = $interface->{interfaceType} eq 'ethernet' ? 'true' : 'false';

			$in->{lookup_tables}->add_interface($name, $interface);
		}

		# Get status of the interface, is it active or not
		my $admin_status = get_value_from_xml('state', $hardware_interface);
		if( not $interface->{adminStatus} and $admin_status =~ /^up|down$/i){
			$interface->{adminStatus} = $admin_status;
		}

		# Get the ethernet data
		if( not $interface->{interfaceEthernet} ) {
			my $interface_ethernet = _parse_interface_ethernet($hardware_interface);
			$interface->{interfaceEthernet} = $interface_ethernet;
		}

		# Get the speed, if it isn't a number throw it away
		my $speed = get_value_from_xml('speed', $hardware_interface);
		if( not defined $interface->{speed} and defined $speed and $speed =~ /^\d+$/ ) {
			$interface->{speed} = $speed * 1000;
		}
	
		my @aggregation_members = $hardware_interface->find_nodes('ae_member/member');
		foreach my $aggregation_member(@aggregation_members){

			my $member = get_value_from_xml('.', $aggregation_member);
			push @{$aggregate_interfaces->{$name}}, $member;
		}
	}

	foreach my $primary_interface(@primary_interfaces){
		my $interface = $in->{lookup_tables}->get_interface($primary_interface);
		add_interface_grouping( $interface, $primary_interface, 'primaryInterface' );
	}

	foreach my $aggregate_parent ( keys %{$aggregate_interfaces} ) {
		my $interface_names = $aggregate_interfaces->{$aggregate_parent};
		foreach my $interface_name ( @{$interface_names} ) {
			my $interface = $in->{lookup_tables}->get_interface($interface_name);
			add_interface_grouping( $interface, $aggregate_parent, 'bondingGroupMember' );
		}
	}

	# Create an management interface for the admin context
	my $mgmt_ip_address = get_value_from_xml( '/response/result/system/ip-address', $in->{show_system_info} );
        my $mgmt_mask = mask_to_bits( get_value_from_xml( '/response/result/system/netmask', $in->{show_system_info}) );
	my $mgmt_interface = {
		adminStatus   => 'up',
		interfaceType => 'virtual',
		name          => 'Management Interface',
		physical      => 'false',
		interfaceIp   => {
			ipConfiguration => {
				ipAddress => $mgmt_ip_address,
				mask      => $mgmt_mask
			}
		}
	};
	push @{$in->{parsed}->{interfaces}}, $mgmt_interface;

	return;
}

# this function takes a hash of XML Twig elements and other data relevant to one interface,
# parses the needed information and returns a hash of data in SED format suitable for xml printing
sub _parse_interface {
	my ( $show_config_entry, $interface_type, $group_characteristic, $lookup, $show_interface_all ) = @_;

	my $name = $show_config_entry->att('name');

	#special case for 'simulated' interfaces : 'vlan', 'loopback', and 'tunnel'
	if (!defined $name) {
		$name = $interface_type;
	}

	$LOGGER->verbose("==Parsing interface: $name");

	my $physical = 'false';

	# Palo alto uses a naming convention for interfaces where a sub-interface has the same name as the
	# primary interface, with a dot and a number tacked on the end.
	my $group_name;
	my $admin_status;
	if ($group_characteristic eq 'primaryInterface') {
		$group_name = $name;
		$admin_status = get_value_from_xml( '/response/result/hw/entry[name="' . $name . '"]/state', $show_interface_all );
	} else {
		($group_name) = $name =~ /^(.*?)\.\d+$/;
		
		#if this interface is a sub interface, use the status of the primary interface
		$admin_status = get_value_from_xml( '/response/result/hw/entry[name="' . $group_name . '"]/state', $show_interface_all );
	}


	my $interface_ethernet;
	my @show_interface_hw_info = $show_interface_all->findnodes( '/response/result/hw/entry[name="' . $name . '"]' );
	#will only be at most one hardware entry for a given interface.
	if (@show_interface_hw_info) {

		#parse the interfaceEthernet information
		$interface_ethernet = _parse_interface_ethernet(shift @show_interface_hw_info);

	} else {
		$LOGGER->verbose("No hardware information found for interface '$name'. Ethernet information not parsed. ");
	}

	#parse the interfaceIp information
	my $interface_ip = _parse_interface_ip($show_config_entry);

	my $parsed_tunnel = _parse_pppoe_tunnel($name, $show_config_entry);

	#palo alto uses a standard interface naming convention
	if( $interface_type eq 'ethernet' and $group_characteristic eq 'primaryInterface' ) {
		$physical = 'true';
	}
	elsif ($interface_type eq 'loopback') {
		$interface_type = 'softwareLoopback';
	}
	elsif ($interface_type eq 'tunnel'){
		$parsed_tunnel = $lookup->get_ipsec_tunnel($name);
	}
	elsif ($interface_type eq 'vlan'){
		$interface_type = 'virtual';
	}

	#parse the interface description
	my $description = get_value_from_xml( 'comment', $show_config_entry );

	#parse the interface speed
	my $speed = get_value_from_xml( '/response/result/hw/entry[name="' . $name . '"]/speed', $show_interface_all );

	if ( ( !defined $speed ) or ( $speed eq 'ukn' ) or ( $speed =~ /n\/a/i ) ) {
		#default to undef so we don't print a SED element. defaulting to zero in some or all cases would also be valid
		$speed = undef;
	} else {
		$speed *= 1000;
	}

	#use the interface ID as the ifIndex as it is unique
	my $if_index = get_value_from_xml( '/response/result/ifnet/entry[name="' . $name . '"]/id', $show_interface_all );

	my $mtu = get_value_from_xml('layer3/mtu', $show_config_entry);

        # Subinterfaces <mtu> element are not under a <layer3> element
	if( not defined $mtu ) {
		$mtu = get_value_from_xml('mtu', $show_config_entry);
	}

	#my $vlan_tag = get_value_from_xml('tag', $show_config_running) if defined $show_config_running;

	#Virtual wire information is parsed in the function _parse_virtual_wires and stored in the lookup table
	my $virtual_wire = $lookup->get_virtual_wire($name);

	my @vlan_membership = $lookup->get_interface_vlan_membership($name);
	my @formatted_vlans;
	foreach my $vlan (@vlan_membership) {
		push @formatted_vlans, {startVlan => $vlan};
	}

	my $vsys_id = get_value_from_xml( '/response/result/ifnet/entry[name="' . $name . '"]/vsys', $show_interface_all );
	if( defined $vsys_id ) {
		$lookup->add_interface_to_vsys( $vsys_id, $name );
	}

	#construct the hash representing the interface
	my $interface = {
		ifIndex           => $if_index,
		name              => $name,
		physical          => $physical,
		interfaceType     => $interface_type,
		adminStatus       => $admin_status,
		description       => $description,
		interfaceEthernet => $interface_ethernet,
		interfaceIp       => $interface_ip,
		interfaceTunnel   => $parsed_tunnel,
		speed             => $speed,
		mtu               => $mtu // $lookup->get_global_mtu(),
	};

	if (defined $group_name) {
		add_interface_grouping( $interface, $group_name, $group_characteristic );
	}
	
	# For 7.2 MR1: make sure to not allow the virtual wire interface grouping to be generated if interface is down--will
	# cause topology to go haywire.
	if (defined $virtual_wire and $admin_status ne 'down') {
		add_interface_grouping( $interface, $virtual_wire, 'proxyAffiliation' );
		push @{$interface->{interfaceBehavior}}, 'bridging';
	}

	my $zone_name = get_value_from_xml( '/response/result/ifnet/entry[name="' . $name . '"]/zone', $show_interface_all );
	if (defined $zone_name) {
		add_interface_grouping( $interface, $zone_name, 'securityZoneMember' );
	}

	if (@formatted_vlans) {
		$interface->{interfaceVlanTrunks} = \@formatted_vlans;
	}

	add_interface_grouping( $interface, 'any', 'securityZoneMember' );

	my $management_profile = get_value_from_xml( 'layer3/interface-management-profile', $show_config_entry );
	if ($management_profile) {
		$lookup->add_interface_to_management_profile($management_profile, $name);
	}

	$lookup->add_interface( $name, $interface );

	return $interface;
}

sub _parse_interface_ethernet {
	my $hw_entry = shift;

	#parse duplex information from the mode value.
	my $mode = get_value_from_xml( 'mode', $hw_entry );
	my $auto_duplex;
	my $auto_speed;

	if ( $mode && $mode =~ /autoneg/ ) {
		$auto_duplex = 'true';
		$auto_speed  = 'true';
	} else {
		$auto_duplex = 'false';
		$auto_speed  = 'false';
	}

	#parse the mac address
	my $mac = strip_mac( get_value_from_xml( 'mac', $hw_entry ) );

	#	my $media_type;

	my $operational_duplex = get_value_from_xml( 'duplex', $hw_entry );
	if ( $operational_duplex =~ /(ukn)|(n\/a)|(unknown)/ ) {
		$operational_duplex = undef;
	}

	my $interface_ethernet = {
		'autoDuplex' => $auto_duplex,
		'autoSpeed'  => $auto_speed,
		'macAddress' => $mac,
	};

	if ( defined $operational_duplex ) {
		$interface_ethernet->{operationalDuplex} = $operational_duplex;
	}

	return $interface_ethernet;
}	#end sub _parse_interface_ethernet


sub _parse_interface_ip {
	my ($show_config_entry) = @_;
	my @ip_configs;

	# IP information can be in one of several different places depending on the interface type. Check them all
	my @entries = $show_config_entry->findnodes('layer3/ip/entry');
	push @entries, $show_config_entry->findnodes('ip/entry');

	foreach my $entry (@entries) {
		my $address = $entry->att('name');

		if( $address =~ /^$IP_REGEX(\/\d+)?$/ ) {
			my $element = parse_cidr_panos($address);
			my $ip_configuration = {
				ipAddress => $element->{host},
				mask	  => $element->{prefix}
			};
			push @ip_configs, $ip_configuration;
		} # if the address is an object reference, PANOS will have resolved it for us in "show interface all", add this address when parsing there
	}

	my $interface_ip;

	# Only attach the list if there are elements
	if ( scalar @ip_configs ) {
		$interface_ip->{ipConfiguration} = \@ip_configs;
	}

	return $interface_ip;
}

sub _parse_pppoe_tunnel{
	my ($interface_name, $interface_entry) = @_;
	my $interface_tunnel;

	my $pppoe = get_value_from_xml('layer3/pppoe/enable', $interface_entry);
	if($pppoe && $pppoe eq 'yes'){

		$interface_tunnel = {
			encapsulationProtocol => 'other',
			tunnelLocalPeering => {tunnelLocalPeeringInterface => $interface_name}
		};

		my $static_address = get_value_from_xml('layer3/pppoe/static-address/ip', $interface_entry);
		if($static_address){
			$interface_tunnel->{tunnelLocalPeering} = {tunnelLocalPeeringIPAddress => $static_address};
		}
	}

	return($interface_tunnel);
}

sub _parse_ipsec_tunnels {
	my ( $tunnel_entries, $gateway_entries, $lookup ) = @_;
	my $parsed_gateways;

	foreach my $entry (@{$gateway_entries}) {
		my $gateway_name = $entry->att('name');
		$parsed_gateways->{$gateway_name}->{tunnelLocalPeeringInterface} = get_value_from_xml('local-address/interface', $entry);
		$parsed_gateways->{$gateway_name}->{tunnelRemotePeerIPAddress} = get_value_from_xml('peer-address/ip', $entry);
	}

	for my $tunnel (@{$tunnel_entries}) {
		my $parsed_tunnel;

		#this function deals exclusively with ipSec tunnels, so default the type accordingly
		$parsed_tunnel->{encapsulationProtocol} = 'ipSec';

		# this function uses 'tunnelLocalPeeringInterface' rather than 'tunnelLocalPeeringIpAddress' as the latter will not always be present

		#ipSec tunnels can be manually configured
		if ($tunnel->first_child_matches('manual-key')) {
			$parsed_tunnel->{tunnelLocalPeering}->{tunnelLocalPeeringInterface} = get_value_from_xml('manual-key/local-address/interface', $tunnel);
			my $peer_ip = get_value_from_xml('manual-key/peer-address/ip', $tunnel);
			# TODO: Cannot support fqdn for peer IP, as object groups are not parsed yet so would require 
			# major rework of adapter ordering, and tunnel peer IP cannot take an object id.
			if( not $peer_ip ) {
				my $fqdn = get_value_from_xml('manual-key/peer-address/fqdn', $tunnel);
			}
			$parsed_tunnel->{tunnelRemotePeerIPAddress} = $peer_ip;
		}

		#ipSec tunnels can be auto configured
		elsif ($tunnel->first_child_matches('auto-key')) {
			if ($tunnel->first_child()->first_child_matches('ike-gateway')) {

				# need to extract the ike gateway name and look up the gateway information
				my $gateway = shift @{$tunnel->findnodes('auto-key/ike-gateway/entry')};
				my $gateway_name = $gateway->att('name');

#				my $gateway_entry = $lookup->get_ike_gateway($gateway_name);
#				my $gateway_entry = shift @{$show_config_running->findnodes('/response/result/config/devices/entry/network/ike/gateway/entry[@name="' .$gateway_name. '"]')};

				$parsed_tunnel->{tunnelLocalPeering}->{tunnelLocalPeeringInterface} = $parsed_gateways->{$gateway_name}->{tunnelLocalPeeringInterface};
				$parsed_tunnel->{tunnelRemotePeerIPAddress} = $parsed_gateways->{$gateway_name}->{tunnelRemotePeerIPAddress};
			}
		}

		#ipSec tunnels can be configured through a global-protect satellite
		elsif ($tunnel->first_child_matches('global-protect-satellite')) {
				$parsed_tunnel->{tunnelLocalPeering}->{tunnelLocalPeeringInterface} = get_value_from_xml('global-protect-satellite/local-address/interface', $tunnel);
				$parsed_tunnel->{tunnelRemotePeerIPAddress} = get_value_from_xml('global-protect-satellite/portal-address', $tunnel);
		}

		#will cause a failed to mergre interfaces parse warning if printed empty
		if( exists( $parsed_tunnel->{tunnelRemotePeerIPAddress} ) && not( $parsed_tunnel->{tunnelRemotePeerIPAddress} ) ) {
			delete( $parsed_tunnel->{tunnelRemotePeerIPAddress} );
		}

		my $tunnel_interface = get_value_from_xml('tunnel-interface', $tunnel);
		$lookup->add_ipsec_tunnel($tunnel_interface, $parsed_tunnel);

	} #end outer for loop
	return;
}

sub _parse_virtual_wires {
	my ( $virtual_wires, $lookup ) = @_;

	foreach my $entry (@{$virtual_wires}) {

		my $name = $entry->att('name');
		my $interface_one = get_value_from_xml('interface1', $entry);
		my $interface_two = get_value_from_xml('interface2', $entry);

		$lookup->add_virtual_wire($name, $interface_one, $interface_two);
	}
	return;
}

sub _parse_management_profile{
	my ($profiles, $lookup) = @_;

	if( defined $profiles ) {
		foreach my $profile(@{$profiles}){

			my $name = $profile->att('name');

			my $network_group;
			my @permitted_ips = $profile->findnodes('permitted-ip/entry');

			foreach my $permitted_ip(@permitted_ips){

				my $name = $permitted_ip->att('name');
				my ($ip, $mask) = split(/\//, $name);

				if(!defined($mask) || $mask == 32){
					push(@{$network_group->{host}}, $ip);
				}else{
					my $subnet = {
						address => $ip,
						mask => $mask
					};
					push(@{$network_group->{subnet}}, $subnet);
				}
			}

			my $service_group;

			my @services = ( qw(http https ssh snmp ping telnet) );
			foreach my $service(@services){

				my $service_permitted = get_value_from_xml($service, $profile);
				if($service_permitted && $service_permitted eq 'yes'){
					
					my $port_expressions = get_port_expression($service);
					foreach my $port_expression(@{$port_expressions}){
						foreach my $expression_type(keys(%{$port_expression})){
							push(@{$service_group->{$expression_type}}, $port_expression->{$expression_type});
						}
					}
				}
			}

			my $response_pages = get_value_from_xml('response-pages', $profile);
			if(defined ($response_pages) && $response_pages eq 'yes'){
				push(@{$service_group->{portRange}}, {
					portStart => 6080,
					portEnd => 6081
				})
			}

			$lookup->add_interface_management_profile($name, $network_group, $service_group);
		}
	}
	return;
}

sub print_interfaces {
	my ( $in, $out ) = @_;

	$out->open_element('interfaces');
	foreach my $interface ( @{ $in->{parsed}->{interfaces} } ) {
		_print_interface($interface, $out);
	}
	$out->close_element('interfaces');

	return;
}

#interface elements are broken out and printed separatly due to the fact that
# the 'interfaceBehavior' element is out of order
sub _print_interface {
	my ($interface, $out) = @_;

	$out->open_element('interface');

	print_if_defined( $interface, 'adminStatus', $out );
	print_if_defined( $interface, 'description', $out );
	print_if_defined( $interface, 'ifIndex', $out );
	print_if_defined( $interface, 'interfaceEthernet', $out );
	print_if_defined( $interface, 'interfaceGrouping', $out );
	print_if_defined( $interface, 'interfaceIp', $out );
	print_if_defined( $interface, 'interfaceTunnel', $out );
	print_if_defined( $interface, 'interfaceBehavior', $out );
	print_if_defined( $interface, 'interfaceType', $out );
	print_if_defined( $interface, 'interfaceVlanTrunks', $out );
	print_if_defined( $interface, 'mtu', $out );
	print_if_defined( $interface, 'name', $out );
	print_if_defined( $interface, 'physical', $out );
	print_if_defined( $interface, 'speed', $out );

	$out->close_element('interface');

	return;
}


sub parse_vlans {
	my ( $in, $out ) = @_;

	my $show_config_running = $in->{'show_config_running'};
	my $lookup = $in->{'lookup_tables'};

	my @parsed_vlans;

	my @vlans = $show_config_running->findnodes('/response/result/config/devices/entry/network/vlan/entry');

	foreach my $vlan (@vlans) {

		my $parsed_vlan;

		$parsed_vlan->{configSource} = 'local';
#		$parsed_vlan->{name} = get_value_from_xml( 'name', $vlan );
		$parsed_vlan->{name} = $vlan->att('name');

		###Vlans in Palo alto don't have numbers directly associated with them, however a number is required by the sed.
		#pull the number from the associated interface if it exists. Otherwise the vlan will be dropped.
		my $virtual_interface = get_value_from_xml( 'virtual-interface/interface', $vlan );
		if (defined $virtual_interface) {
		  ( $parsed_vlan->{number} ) = $virtual_interface =~ /vlan\.(\d+)/;
		  $lookup->add_vlan_member($parsed_vlan->{name}, $virtual_interface);
		}

		$parsed_vlan->{enabled} = 'true';

		#parse the interface members of this vlan
		my @members		= $vlan->findnodes('interface/member');
		my @parsed_members = ();
		foreach my $member (@members) {
			my $interface_member = get_value_from_xml( '.', $member );
			push @parsed_members, $interface_member;
			$lookup->add_vlan_member($parsed_vlan->{name}, $interface_member);
		}

		if (@parsed_members) {
			$parsed_vlan->{interfaceMember} = \@parsed_members;
		}

		#Only save the vlan if it has a valid number
		if ( defined $parsed_vlan->{number} ) {
			push @parsed_vlans, $parsed_vlan;
		} else {
			$LOGGER->debug( "Parsing of VLAN '" . $parsed_vlan->{name} . "' failed. VLAN has no number. This is likely due to it having no associated vlan interface." );
		}
	}

	if (@parsed_vlans) {
		$in->{parsed}->{vlans}->{vlan} = \@parsed_vlans;
	}

	return;
}

sub print_vlans {
	my ( $in, $out ) = @_;

	print_if_defined($in->{parsed}, 'vlans', $out);
#	$out->print_element( 'vlans', $in->{parsed}->{vlans} );

	return;
}

# This sets up parsing of the <shared> definitions of the configuration XML from the device.
sub parse_shared_elements 
{
	my ($in, $out) = @_;
	my $show_config_running = $in->{'show_config_running'};

	# There is only one of these
	my $sharedSystemElements = $show_config_running->findnodes('/response/result/config/shared');
	if ($sharedSystemElements)
	{
		my $parsedSharedElementSet->{objectGroups} = _parse_object_groups($in, $sharedSystemElements->[0]);
		if ($parsedSharedElementSet) {
			$in->{parsed}->{sharedElements} = $parsedSharedElementSet; 
		}
	} else {
		$LOGGER->debug("Parsing of shared definitions bypassed: no applicable shared elements found");	
	}

	return;
}

# This function extracts the VSystem information from the running config (and anywhere else we find it)
# and calls sub parsers to create subelement entities to include in the SED.
sub parse_vsystems {
	my ($in) = @_;
	my @parsed_vsystems;

	my $show_config_running = $in->{'show_config_running'};
	my @virtual_systems = (
		$show_config_running->findnodes('/response/result/config/devices/entry/vsys/entry'),
		$show_config_running->findnodes('/response/result/config/devices/entry/network/shared-gateway/entry')
	);

	foreach my $vsys (@virtual_systems) {

		my $parsed_vsys;
		my $name = $vsys->att('name');
		$parsed_vsys->{'core:systemName'} = $name;
		
		#virtual systems are essentially virtual firewalls
		$parsed_vsys->{'core:deviceType'} = 'Firewall';
		$parsed_vsys->{'subelementType'} = 'virtual';

		#lookup the interfaces that are in this virtual system
		my @imported_interfaces = $vsys->findnodes('import/network/interface/member');
		my @visible_vsys = $vsys->findnodes('import/visible-vsys/member');
		my $interfaces = _import_interfaces( $name, \@imported_interfaces, \@visible_vsys, $in->{lookup_tables} );
		$parsed_vsys->{interfaces} = $interfaces;
		
		# static routes
		my $vsysCount = @virtual_systems;
		my @staticRoutes;
		if ($vsysCount == 1 && $in->{parsed}->{staticRoutes}) {
			@staticRoutes = $in->{parsed}->{staticRoutes}; # for single context device use all static routes
		} else{
			@staticRoutes = parse_static_routes($vsys, $in->{lookup_tables}, $vsysCount);
		}
		$parsed_vsys->{staticRoutes} = { staticRoute => \@staticRoutes } if @staticRoutes;
		
		# Parse the applications defined on this vsys
		my $parsed_applications = parse_applications($vsys);
		$parsed_vsys->{applications} = { application => $parsed_applications } if $parsed_applications;

		# Parse the object groups in this vsys
		$parsed_vsys->{objectGroups} = _parse_object_groups( $in, $vsys, $name );

		# Parse the object groups Panorama has pushed to this vsys
		my @panorama_configs = $in->{show_config_panorama}->findnodes("/response/result/panorama/vsys/entry[\@name='$name']");
		if( @panorama_configs ) {
			my $parsed_panorama_objects = _parse_object_groups( $in, $panorama_configs[0], $name );
			push @{$parsed_vsys->{objectGroups}->{networkGroup}}, @{$parsed_panorama_objects->{networkGroup}};
			push @{$parsed_vsys->{objectGroups}->{serviceGroup}}, @{$parsed_panorama_objects->{serviceGroup}};
			push @{$parsed_vsys->{objectGroups}->{applicationGroup}}, @{$parsed_panorama_objects->{applicationGroup}};
		}

		# Parse the rules in this vsys
		$parsed_vsys->{filterLists} = parse_acls( $vsys, $name, $in->{lookup_tables}, $in->{show_config_panorama});
		
		# Add any needed application-default service groups to object groups
		my $adp_service_groups = $in->{lookup_tables}->get_default_port_service_groups();
		push @{$parsed_vsys->{objectGroups}->{serviceGroup}}, @{$adp_service_groups};

		my $nat_filter_list = parse_nat( $in, $vsys, $name );
		if( $nat_filter_list ) {
			$parsed_vsys->{natFilterLists} = $nat_filter_list;
		}

		push @parsed_vsystems, $parsed_vsys;
	}

	#There is always at least one vsys
	$in->{parsed}->{subelements}->{subelement} = \@parsed_vsystems;

	return;
}

sub parse_static_routes {
	my $vsys = shift;
	my $lookup = shift;
	
	my @static_routes;
	my @imported_virtual_routers = $vsys->findnodes('import/network/virtual-router/member');
	foreach my $member (@imported_virtual_routers){
		my $vr_name = get_value_from_xml( '.', $member );
		my $virtual_router = $lookup->get_virtual_router($vr_name);
		if ( $virtual_router ) {
			push @static_routes, @{$virtual_router->{staticRoutes}};
		} else {
			$LOGGER->debug("WARNING: Failed to import static routes from virtual router '$vr_name' into a virtual device. This virtual router does not exist in the lookup table.");
		}
	}
	return @static_routes;
}

#this function takes a vsys entity and parses the object group information it contains
sub _parse_object_groups {
	my ($in, $vsys, $vsys_name) = @_;

	my $parsed_object_groups;

	$parsed_object_groups->{networkGroup} = _parse_network_groups( $in, $vsys );
	$parsed_object_groups->{serviceGroup} = _parse_service_groups( $in, $vsys );
	$parsed_object_groups->{applicationGroup} = _parse_application_groups( $in, $vsys );	
	my $parsed_application_filters = _parse_application_filters( $in, $vsys );
        push @{ $parsed_object_groups->{applicationGroup} }, @{ $parsed_application_filters };

	# Don't add these to the object groups hash as they are not displayed at present. Also, not present for shared elements.
	my ($zones) = $vsys->findnodes('zone');
	if( $zones ) {
		_parse_security_zones( $zones, $vsys_name, $in->{lookup_tables} );
	}
	
	# Parse user groups to be used in future
	#$LOGGER->debug('==Parsing user group information');
	#$parsed_object_groups->{userGroup} = parse_user_groups ($vsys, $lookup, $show_user_group_name);

	return $parsed_object_groups;
}

#this function takes an address entity and an address group entity parses them and returns
#a list of hashes structured as SED networkGroup elements
sub _parse_network_groups {
	my ( $in, $vsys ) = @_;
	my $vsys_id = $vsys->att('name');
        my $lookup = $in->{lookup_tables};
        my $shared_network_groups = $in->{parsed}->{sharedElements};
        my $show_dynamic_address_groups = $in->{show_dynamic_address_groups};
        my $request_dynamic_block_list = $in->{request_dynamic_block_list};
	my $request_fqdn = $in->{request_fqdn};

	my @parsed_network_groups = ();

	# We collect any FQDNs and later drop them if they are unresolved
	my %fqdns;
	my @address_entries = $vsys->findnodes('address/entry');
	foreach my $entry (@address_entries) {

		my $id = $entry->att('name');
		my $parsed_network_group = { 'id' => $id, };

		my $description = get_value_from_xml( 'description', $entry );
		if ( defined $description ) {	#Only save the description if it is not empty
			$parsed_network_group->{description} = $description;
		}

		#network groups can be defined with netmasks or ranges
		if ( $entry->exists('ip-netmask') ) {
			my $cidr = get_value_from_xml( 'ip-netmask', $entry );

			my $element = parse_cidr_panos($cidr);
			$parsed_network_group->{subnet}->{address} = $element->{host};
			$parsed_network_group->{subnet}->{mask}	= $element->{prefix};

		} elsif ( $entry->exists('ip-range') ) {

			my $range = get_value_from_xml( 'ip-range', $entry );

			my ( $start, $end ) = $range =~ /(.*)-(.*)/;

			$parsed_network_group->{range}->{startAddress} = $start;
			$parsed_network_group->{range}->{endAddress}   = $end;
		} elsif( $entry->exists('fqdn') ) {

			my $fqdn = get_value_from_xml('fqdn', $entry);
			my @hosts;

			# The api request to show a specific FQDN object's resolution is only available
			# from PAN-OS v8.1 and greater. Also not available via Panorama.
			if( $in->{version} gt '8.1' and not $in->{is_proxy_backup} ) {
				my $response = &{ $request_fqdn }($vsys_id, $fqdn);

				if( $response ) {
					@hosts = $response->findnodes('response/result/fqdn/entry/ip/member');
					for my $host_node ( @hosts ) {
						my $host = get_value_from_xml('.', $host_node);
						push @{ $parsed_network_group->{host} }, $host;
					}
				}
			}

			if( not @hosts ) {
				$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Dropping unresolved FQDN '$fqdn'.");
				$fqdns{$fqdn} = 1;
			}
		}

		push @parsed_network_groups, $parsed_network_group;
	}

	# We collect dynamic address group names and descs from vsys running config, then populate with addresses after this loop
	my %dynamic_address_groups;
	my @address_group_entries = $vsys->findnodes('address-group/entry');
	foreach my $entry (@address_group_entries) {
		my $id = $entry->att('name');
		my $parsed_network_group = { 'id' => $id, };

		my $description = get_value_from_xml( 'description', $entry );
		if ( defined $description ) {
			$parsed_network_group->{description} = $description;
		}

		$parsed_network_group->{objectGroupReference} = ();

		# Dynamic address groups addresses are parsed later
		if( $entry->exists('dynamic') ) {
			$dynamic_address_groups{$id} = $parsed_network_group;
		} else {

			# Two XPaths used for pre and post PAN-OS 6.0
			my @members = ();
			foreach my $xpath('member', 'static/member'){
				my $member_nodes = $entry->findnodes($xpath);
				while($member_nodes->size() > 0){
					push @members, $member_nodes->shift();
				}
			}

			foreach my $member (@members) {
				my $reference = get_value_from_xml( '.', $member );
				if( defined $fqdns{$reference} ) {
					$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Dropping unresolved nested FQDN '$reference'.");
				} else {
					push @{ $parsed_network_group->{objectGroupReference} }, $reference;
				}
			}

			push @parsed_network_groups, $parsed_network_group;
		}
	}

	# Only do this if we have the response with the dynamic address group members
	# and if this is not the shared parser, but the vsys parser
	my $parsed_dynamic_address_groups = _parse_dynamic_address_groups($vsys, $show_dynamic_address_groups, \%dynamic_address_groups);
	push @parsed_network_groups, @{ $parsed_dynamic_address_groups };

	my $profiles = $lookup->get_interface_management_profiles();
	foreach my $profile(@{$profiles}){

		my $parsed_network_group = $profile->{network_group};

		my $id = $profile->{name};
		$parsed_network_group->{id} = "Interface management profile $id network group";
		
		push(@parsed_network_groups, $parsed_network_group);
	}

	my $external_list_groups = _parse_external_lists($vsys, $request_dynamic_block_list, $in->{version});
	push @parsed_network_groups, @{ $external_list_groups } if $external_list_groups;

	# Find all referenced network groups in this virtual context's filters
	my %referenced_network_groups;
	my @rulebase = $vsys->findnodes('rulebase/security/rules/entry');
	foreach my $rule(@rulebase){
		my @addresses = ($rule->findnodes('source/member'), $rule->findnodes('destination/member'));
		foreach my $address(@addresses){
			my $address_value = get_value_from_xml('.', $address);
			if($address_value ne 'any' && $address_value !~ /^\d+\.\d+\.\d+\.\d+(\/\d+)?$/){
				$referenced_network_groups{$address_value} = 1;
			}
		}
	}

	# Find all referenced network groups in this virtual context's NAT filters
	my @nat_rulebase = $vsys->findnodes('rulebase/nat/rules/entry');
	foreach my $nat_rule(@nat_rulebase){
		my @addresses = (
			$nat_rule->findnodes('source/member'),
			$nat_rule->findnodes('destination/member'),
			$nat_rule->findnodes('destination-translation/translated-address'),
			$nat_rule->findnodes('source-translation/static-ip/translated-address'),
			$nat_rule->findnodes('source-translation/dynamic-ip/translated-address/member'),
			$nat_rule->findnodes('source-translation/dynamic-ip-and-port/translated-address/member')
		);
		foreach my $address(@addresses){
			my $address_value = get_value_from_xml('.', $address);
			if($address_value ne 'any' && $address_value !~ /^\d+\.\d+\.\d+\.\d+(\/\d+)?$/){
				$referenced_network_groups{$address_value} = 1;
			}
		}
	}

	# Find any nested shared network groups
	foreach my $shared_network_group(@{$shared_network_groups->{objectGroups}->{networkGroup}}){
		if($shared_network_group->{objectGroupReference}){
			foreach my $object_group_ref(@{$shared_network_group->{objectGroupReference}}){
				$referenced_network_groups{$object_group_ref} = 1;
			}
		}
	}

	# Same as above but look inside network groups defined for this virtual context
	foreach my $parsed_network_group(@parsed_network_groups){
		if($parsed_network_group->{objectGroupReference}){
			foreach my $object_group_ref(@{$parsed_network_group->{objectGroupReference}}){
				$referenced_network_groups{$object_group_ref} = 1;
			}
		}
	}

	# Add any we found that are used
	foreach my $shared_network_group(@{$shared_network_groups->{objectGroups}->{networkGroup}}){
		if($referenced_network_groups{$shared_network_group->{id}}){
			push(@parsed_network_groups, $shared_network_group);
		}
	}
	
	# Add parsed network groups to lookup
	foreach my $group (@parsed_network_groups) {
		$lookup->add_network_group( $group->{id}, $group );
	}

	return \@parsed_network_groups;
}

sub _parse_dynamic_address_groups {
	my ($vsys, $show_dynamic_address_groups, $dynamic_address_groups_ref) = @_;
	my %dynamic_address_groups = %{ $dynamic_address_groups_ref };
	my $vsys_id = $vsys->att('name');

	my @parsed_dynamic_address_groups = ();

	if( defined $show_dynamic_address_groups && defined $vsys_id ) {
		my @dynamic_address_groups = $show_dynamic_address_groups->findnodes('/response/result/dyn-addr-grp/entry');
		foreach my $entry (@dynamic_address_groups) {
			my $group_id = get_value_from_xml( 'group-name', $entry );
			my $group_vsys_id = get_value_from_xml( 'vsys', $entry );
 
			if( $group_vsys_id eq $vsys_id ) {
				if( defined $dynamic_address_groups{$group_id} ) {
					my @members = $entry->findnodes('member-list/entry');
					foreach my $member (@members) {
						my $name = $member->att('name');
						my $type = $member->att('type');
						if( $type eq 'registered-ip' ) {
							push @{ $dynamic_address_groups{$group_id}->{subnet} }, { address => $name, mask => 32 };
						} elsif( $type eq 'address-object' ) {
							push @{ $dynamic_address_groups{$group_id}->{objectGroupReference} }, $name;
						} else {
							$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Dynamic address group '$group_id' member '$name' has an unknown type '$type'");
						}
					}

					push @parsed_dynamic_address_groups, $dynamic_address_groups{$group_id};
				} else {
					$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Dynamic address group '$group_id' was not in the running-config but was in the dynamic address groups response");
				}
			}
		}
	} else {
		foreach my $id ( keys %dynamic_address_groups ) {
			$LOGGER->warning($NETWORK_OBJECT_DROPPED, "No response for dynamic address groups. Dropping dynamic address group '$id' found in the running-config");
		}
	}

	return \@parsed_dynamic_address_groups;
}

sub _parse_external_lists {
	my ($vsys, $request_dynamic_block_list, $version) = @_;
	my $network_groups;

	my @external_lists = $vsys->findnodes('external-list/entry');
	for my $external_list ( @external_lists ) {

		# In PAN-OS v7.1 and beyond the external lists are stored under a new element called type.
		# We only want the lists that refer to IP addresses. We do not want URL or domain lists, skip these.
		my $type = '';
		if( $external_list->has_child('type') ) {
			my $type_twig = $external_list->first_child('type');
			if( $type_twig ) {
				$type_twig = $type_twig->child(0);
				$type = $type_twig->is_elt() ? $type_twig->name() : $type_twig->text();
				if( $type ne 'ip' and $type ne 'predefined-ip' ) {
					next;
				}
			}
		}

		my $name = $external_list->att('name');
		my $model = { id => $name };

		my $description = get_value_from_xml('description', $external_list);
		$model->{description} = $description if $description;

		# Request the addresses in this dynamic block list
		my $dbl = &{ $request_dynamic_block_list }($name, $type, $version);

		# There are two syntaxes for responses of the addresses
		my @members;
		my $line = get_value_from_xml('response/msg/line', $dbl);
		if( $line ) {
			@members = split /^/, $line;
		} else {
			my @member_nodes = $dbl->findnodes('response/result/external-list/valid-members/member');
			for my $member_node ( @member_nodes ) {
				push @members, get_value_from_xml('.', $member_node);
			}
		}

		# Add to a set to remove duplicates and allow for future exclusion support
		my $cidr_set = Net::CIDR::Set->new({type => 'ipv4'});

		MEMBER:
		for my $member ( @members ) {
			if( $member =~ /^\s*($IP_REGEX)(\/(\d{1,2})|-($IP_REGEX))?(\s|$)/ ) {
				my $ip = $1;
				my $mask = $3 // 32;

				my $range_end = $4;
				if( $range_end ) {
					my @ip_octals = split /\./, $ip;
					my @range_octals = split/\./, $range_end;
					for( my $i = 0; $i < scalar @ip_octals; ++$i ) {
						if( $ip_octals[$i] > $range_octals[$i] ) {
							next MEMBER;
						}
					}
					$cidr_set->add("$ip-$range_end");
				} else {
					$cidr_set->add("$ip/$mask");
				}
			}
		}

		# Introduced in PANOS v8.1, there can be manually defined exceptions to the list
		my $except_set = Net::CIDR::Set->new({type => 'ipv4'});
		my @exceptions = $external_list->findnodes("type/$type/exception-list/member");
		for my $exception ( @exceptions ) {
			my $value = get_value_from_xml('.', $exception);
			$except_set->add($value);
		}
		my $diff_set = $cidr_set->diff($except_set);

		#  TODO: Do not use as_range_array until Net::CIDR::Set is 0.13 or above
		my @diffs = $diff_set->as_cidr_array();

		my (@hosts, @subnets, @ranges);
		for my $address ( @diffs ) {
			if( $address =~ /(.+)-(.+)/ ) {
				push @ranges, { startAddress => $1, endAddress => $2 };
			} else {
				my ($address, $mask) = split /\//, $address;
				if( $mask ) {
					push @subnets, { address => $address, mask => $mask };
				} else {
					push @hosts, $address;
				}
			}
		}

		$model->{host} = \@hosts if @hosts;
		$model->{subnet} = \@subnets if @subnets;
		$model->{range} = \@ranges if @ranges;

		push @{ $network_groups }, $model;
	}

	return $network_groups;
}

sub _parse_service_groups {
	my ( $in, $vsys ) = @_;
        my $lookup = $in->{lookup_tables};
        my $shared_service_groups = $in->{parsed}->{sharedElements};
        my $predefined_services = $in->{show_predefined_service_groups};

	my @parsed_services;

	# The list of all services includes predefined services and user defined services
	my @service_entries = $vsys->findnodes('service/entry');
	
	if (defined $predefined_services) {
	   push @service_entries, $predefined_services->findnodes('/response/result/service/entry');
	}else{
		$LOGGER->debug('Predefined service groups not parsed because the command failed.');
	}

	# Palo Alto uses the term service to mean a set of source/destination ports on a particular protocol
	# We translate this as a service group
	# this should not be confused with the Palo Alto usage of service group which is dealt with later
	foreach my $entry (@service_entries) {
		my $id = $entry->att('name');
		my $description = get_value_from_xml( 'description', $entry );

		my $protocol;
		if ( $entry->exists('protocol/tcp') ) {
			$protocol = 'tcp';
		} elsif ( $entry->exists('protocol/udp') ) {
			$protocol = 'udp';
		}

		### Since the SED can not represent a service group with a destinction between source and destination ports,
		### We have to split each service into source and destination components.

		# This section parses the destination portion of a service group
		if ( $entry->exists("protocol/$protocol/port") ) {
			my $parsed_service_destination = { 'id' => $id . '_destination', };

			# Only save the description if it is not blank
			if ( defined $description ) {
				$parsed_service_destination->{description} = $description;
			}

			my $ports = get_value_from_xml( "protocol/$protocol/port", $entry );

			# Break the port element into a set of ports and port ranges
			my ( $port_expressions, $port_ranges ) = unwind_port_expression($ports);

			foreach my $expression ( @{$port_expressions} ) {
				push @{ $parsed_service_destination->{portExpression} }, { 'port' => $expression, 'operator' => 'eq', 'protocol' => $protocol };
			}
			foreach my $range ( @{$port_ranges} ) {
				my ( $start, $end ) = $range =~ /(.*)-(.*)/;
				push @{ $parsed_service_destination->{portRange} }, { 'portStart' => $start, 'portEnd' => $end, 'protocol' => $protocol };
			}

			push @parsed_services, $parsed_service_destination;
		}

		# This section parses the source section of a service group
		if ( $entry->exists("protocol/$protocol/source-port") ) {
			my $parsed_service_source = { 'id' => $id . '_source', };

			# Only save the description if it is not blank
			if ( defined $description ) {
				$parsed_service_source->{description} = $description;
			}

			my $ports = get_value_from_xml( "protocol/$protocol/source-port", $entry );

			# Break the port element into a set of ports and port ranges
			my ( $port_expressions, $port_ranges ) = unwind_port_expression($ports);

			foreach my $expression ( @{$port_expressions} ) {
				push @{ $parsed_service_source->{portExpression} }, { 'port' => $expression, 'operator' => 'eq', 'protocol' => $protocol };
			}

			foreach my $range ( @{$port_ranges} ) {
				my ( $start, $end ) = $range =~ /(.*)-(.*)/;
				push @{ $parsed_service_source->{portRange} }, { 'portStart' => $start, 'portEnd' => $end, 'protocol' => $protocol };
			}

			push @parsed_services, $parsed_service_source;
		}

	}

	# Palo Alto uses the term service-group to mean a set of references to 'services' or other service groups
	# We create source and destination versions of each of these groups and then trim all of the empty groups
	# and references to empty groups.
	my @service_group_entries = $vsys->findnodes('service-group/entry');

	my @parsed_service_groups;
	foreach my $entry (@service_group_entries) {
		my $id = $entry->att('name');

		# Break the service group up into source and destination components by default. We will trim the list later
		my $parsed_service_group_source	     = { 'id' => $id . '_source', };
		my $parsed_service_group_destination = { 'id' => $id . '_destination', };

		$parsed_service_group_source->{objectGroupReference}	  = ();
		$parsed_service_group_destination->{objectGroupReference} = ();

		# Set up the objectGroupReferences
		# Two XPaths for pre and post PAN-OS 6.0
		my @members = ();
		foreach my $xpath('member', 'members/member'){
			my $member_nodes = $entry->findnodes($xpath);
			while($member_nodes->size() > 0){
				push(@members, $member_nodes->shift());
			}
		}

		foreach my $member (@members) {

			my $reference = get_value_from_xml( '.', $member );
			push @{ $parsed_service_group_source->{objectGroupReference} },	  $reference . '_source';
			push @{ $parsed_service_group_destination->{objectGroupReference} }, $reference . '_destination';
		}

		push @parsed_service_groups, $parsed_service_group_source;
		push @parsed_service_groups, $parsed_service_group_destination;
	}

	# Trim the service group list. See TrimServiceGroups.pm for details. +++wayne note: I'm not buying this 
	# is entirely necessary, but if so, it needs to take into account the multi-vsys/shared scope defs.
	my $parsed_service_groups = trim_service_groups( \@parsed_services, \@parsed_service_groups, $shared_service_groups );

	# Register the service groups in the lookup table to they can be used later
	foreach my $group ( @{$parsed_service_groups} ) {
		$lookup->add_service_group( $group->{id}, $group );
	}

	my $profiles = $lookup->get_interface_management_profiles();
	foreach my $profile(@{$profiles}){

		my $parsed_service_group = $profile->{service_group};

		my $id = $profile->{name};
		$parsed_service_group->{id} = "Interface management profile $id service group";

		push(@{$parsed_service_groups}, $parsed_service_group);
	}

	foreach my $shared_service_group(@{$shared_service_groups->{objectGroups}->{serviceGroup}}){
		push(@{$parsed_service_groups}, $shared_service_group);
	}

	#return the list of all parsed and decomposed services and all parsed service groups
	return $parsed_service_groups;
}

# Palo Alto uses the concept of security zones when creating ACLs. A security zone is a named set of interfaces.
# This subroutine creates these lists of interface names and stores them under the zone name in the lookup.
sub _parse_security_zones {
	my ( $config, $vsys_name, $lookup ) = @_;

	my @entries = $config->findnodes('entry');
	for my $entry ( @entries ) {
		my $zone_name = $entry->att('name');
		my @interfaces = ();

		# collect interfaces defined in the zone
		my @interface_types = qw( tap virtual-wire layer2 layer3 );
		for my $type ( @interface_types ) {
			my @members = $entry->findnodes( "network/$type/member" );
			for my $member ( @members ) {
				push @interfaces, get_value_from_xml( '.', $member );
			}
		}

=cut
		# an external zone means all the interfaces from another virtual system
		my @members = $entry->findnodes('network/external/member');
		for my $member ( @members ) {
			my $external_vsys = get_value_from_xml( '.', $member );
			my $external_interfaces = $lookup->get_interfaces_by_vsys( $external_vsys );
			if( $external_interfaces ) {
				push @interfaces, @{ $external_interfaces };

				# Retroavtively add this zone to the interface 
				for my $external_interface ( @{ $external_interfaces } ) {
					my $interface_model = $lookup->get_interface($external_interface);
					add_interface_grouping( $interface_model, $zone_name, 'securityZoneMember' );
				}
			} else {
				$LOGGER->debug("Could not get interfaces for vsys $external_vsys");
			}
		}
=cut
		$lookup->add_zone_to_vsys( $zone_name, $vsys_name, \@interfaces );
	}
	return;
}

sub _parse_application_groups {
	my ( $in, $vsys ) = @_;
	my $lookup = $in->{lookup_tables};

	my @parsed_application_groups = ();

	my @application_groups = $vsys->findnodes('application-group/entry');
	for my $application_group (@application_groups) {

		my $name = $application_group->att('name');
		my $parsed_application_group = {
			id => $name,
			description => "Application group for application group $name"
		};

		my @members = $application_group->findnodes('member');
		for my $member (@members) {
			my $member_name = get_value_from_xml('.', $member);

			if ($lookup->is_application_group_defined($member_name)) {
				push @{$parsed_application_group->{objectGroupReference}}, $member_name;
			} else {
				push @{$parsed_application_group->{application}}, $member_name;
			}
		}

		$lookup->add_application_group($name, $parsed_application_group);

		push @parsed_application_groups, $parsed_application_group;
		
	}

	return \@parsed_application_groups;	
}

sub _parse_application_filters{
	my ( $in, $vsys ) = @_;
	my $lookup = $in->{lookup_tables};

	my @parsed_application_filters = ();

	my @classifiers = qw(
		risk
		category
		subcategory
		technology
		evasive-behavior
		consume-big-bandwidth
		used-by-malware
		able-to-transfer-file
		has-known-vulnerability
		tunnel-other-application
		prone-to-misuse
	);

	my @application_filters = $vsys->findnodes('application-filter/entry');
	for my $application_filter ( @application_filters ) {
		my $name = $application_filter->att('name');
		my $search_criteria;

		for my $filter_type(@classifiers){

			my @members = $application_filter->findnodes("$filter_type/member");
			for my $member(@members){
				my $member_value = get_value_from_xml('.', $member);
				push(@{$search_criteria->{$filter_type}}, $member_value);
			}
		}

		my $parsed_application_filter = _parse_application_filter_group($name, $search_criteria, $lookup);

		$lookup->add_application_filter($name, $parsed_application_filter);

		push @parsed_application_filters, $parsed_application_filter;
	}

	return \@parsed_application_filters;
}

sub _parse_application_filter_group{
	my ($name, $search_criteria, $lookup) = @_;
	my $parsed_application_filter = {
		id => $name,
		description => "Application group for application filter $name"
	};

	my $applications = $lookup->get_applications();
	my @application_keys = keys(%{$applications});

	foreach my $search_criterion_key(keys(%{$search_criteria})){
		my $search_criterion = $search_criteria->{$search_criterion_key};
		foreach my $search_criterion_value(@$search_criterion){

			my @new_application_keys = ();

			foreach my $application_key(@application_keys){
				my $application = $applications->{$application_key};

				foreach my $criteria_key(keys(%{$application->{criteria}})){

					if($criteria_key eq $search_criterion_key){
						my $criteria = $application->{criteria}->{$criteria_key};

						foreach my $criteria_value(@$criteria){
							if($criteria_value eq $search_criterion_value){
								push(@new_application_keys, $application_key);
							}
						}
					}
				}
			}
			@application_keys = @new_application_keys;
		}
	}

	$parsed_application_filter->{application} = \@application_keys;

	return $parsed_application_filter;
}

sub _print_application_groups {
	my ( $system, $out ) = @_;
	
	if (!defined $system->{applicationGroup}) {
		return;
	}
	
	foreach my $group (@{$system->{applicationGroup}}) {
		$out->open_element('applicationGroup');
		print_if_defined($group, 'description', $out);
		print_if_defined($group, 'id', $out);
		print_if_defined($group, 'objectGroupReference', $out);
		print_if_defined($group, 'application', $out);
		$out->close_element('applicationGroup');
	}

	return;
}



sub _print_object_groups {
	my ( $system, $out ) = @_;
	
	if (!defined $system->{objectGroups}) {
		return;
	}
	
	$out->open_element('objectGroups');
	
	_print_application_groups($system->{objectGroups}, $out);
	
	#print network groups
	if($system->{objectGroups}->{networkGroup}){

		foreach my $network_group(@{$system->{objectGroups}->{networkGroup}}){
			$out->open_element('networkGroup');

			print_if_defined($network_group, 'description', $out);
			print_if_defined($network_group, 'id', $out);
			print_if_defined($network_group, 'objectGroupReference', $out);
			print_if_defined($network_group, 'host', $out);
			print_if_defined($network_group, 'range', $out);
			print_if_defined($network_group, 'subnet', $out);

			$out->close_element('networkGroup');
		}
	}
	
	print_if_defined($system->{objectGroups}, 'serviceGroup', $out);

	# won't be any, as the parsing is commented out	
	if($system->{objectGroups}->{userGroup}){

		foreach my $user_group ( @{$system->{objectGroups}->{userGroup}} ) {
			$out->open_element('userGroup');

			print_if_defined($user_group, 'id', $out);
			foreach my $user (@{$user_group->{user}}){
				$out->print_element('user', $user);
			}
			$out->close_element('userGroup');
		}
	}
	$out->close_element('objectGroups');
	return;
}

sub parse_virtual_routers {
	my ($in) = @_;

	my $show_config_running = $in->{'show_config_running'};
	my $lookup = $in->{lookup_tables};
	my @all_static_routes;

	#get a list of all virtual routers on the device
	my @routers = $show_config_running->findnodes('/response/result/config/devices/entry/network/virtual-router/entry');

	foreach my $router (@routers) {
		my $parsed_router;
		my $name = $router->att('name');
		
		#Extract static route elements
		$LOGGER->debug( "Parsing static route information for virtual router '$name'" );
		my @static_routes = $router->findnodes('routing-table/ip/static-route/entry');

		#parse route information for each route
		foreach my $static_route (@static_routes) {
			my $parsedStaticRoute = _parse_static_route($static_route);
			push @{ $parsed_router->{staticRoutes} }, $parsedStaticRoute if $parsedStaticRoute;
			push @all_static_routes, $parsedStaticRoute if $parsedStaticRoute;
			
		}
		$lookup->add_virtual_router($name, $parsed_router);
		
	}
	$in->{parsed}->{staticRoutes} = \@all_static_routes;
	return;
}

sub _import_interfaces {
	my ( $vsys_name, $interfaces, $visible_vsys, $lookup ) = @_;
	my $imported_interfaces;

	for my $member ( @{ $interfaces } ) {
		my $if_name = get_value_from_xml( '.', $member );
		my $interface = $lookup->get_interface( $if_name );
		
		# The interface may not be included in our lookup for this vsys add it to make sure.
		# This happens with shared gateways, as they have a seemingly random vsys id in "show interface all" 
		my $is_shared_gateway = $vsys_name =~ /^sg\d+$/;
		my ($vsys_id) = $vsys_name =~ /^\D+(\d+)$/;
		$lookup->add_interface_to_vsys( $vsys_id, $interface, $is_shared_gateway );

		push @{ $imported_interfaces->{interface} }, $interface;
	}

	return $imported_interfaces;
}

sub _parse_static_route {
	my $route = shift;

	my $parsed_static_route;
	
	my $name = $route->att('name');

	$parsed_static_route->{defaultGateway} = 'false';

	my $destination = get_value_from_xml( 'destination', $route );

	my ( $address, $mask ) = $destination =~ /(\d+\.\d+\.\d+\.\d+)\/(\d+)/;
	
	if ($address eq "0.0.0.0"){
		$parsed_static_route->{defaultGateway} = "true";
	}
	
	$parsed_static_route->{destinationAddress} = $address;
	$parsed_static_route->{destinationMask}	= $mask;

	my @discard = $route->findnodes('nexthop/discard');
	my @next_vr = $route->findnodes('nexthop/next-vr');
	if (@discard) {
		$parsed_static_route->{discard} = 'true';
	} elsif (@next_vr) {
		$parsed_static_route->{discard} = 'false';
		$LOGGER->debug("WARNING: Dropping static route '$name'. Next hop virtual routers are not supported by QRM.");
		return;
	} else {
		$parsed_static_route->{discard} = 'false';

		my $gateway_address = get_value_from_xml( 'nexthop/ip-address', $route );
		if ( defined $gateway_address ) {
			$gateway_address =~ s/\/\d+//;
			$parsed_static_route->{gatewayAddress} = $gateway_address;
		}
	}

	my $interface = get_value_from_xml( 'interface', $route );
	if ( defined $interface ) {
		$parsed_static_route->{interface} = $interface;
	} else {
		#interface is required in the SED, but routes that discard packets don't have interfaces
		$parsed_static_route->{interface} = 'null';
	}

	#get the route metric
	my $metric = get_value_from_xml( 'metric', $route );
	if ( defined $metric ) {
		$parsed_static_route->{routeMetric} = $metric;
	}

	#get the admin distance
	my $preference = get_value_from_xml( 'admin-dist', $route );
	if ( defined $preference ) {
		$parsed_static_route->{routePreference} = $preference;
	}

	return $parsed_static_route;
}

sub print_shared_object_groups {
	my ($in, $out) = @_;
	print_if_defined($in->{parsed}->{sharedElements}, 'objectGroups', $out);
	return;
}

sub print_sub_elements {
	my ( $in, $out, $qradarVersion ) = @_;

	$out->open_element('subelements');

	foreach my $system (@{ $in->{parsed}->{subelements}->{subelement} }) {

		$out->open_element('subelement');

		print_if_defined( $system, 'core:systemName', $out );
		print_if_defined( $system, 'core:deviceType', $out );
		print_if_defined( $system, 'applications', $out );

		_print_object_groups( $system, $out );

		print_filter_lists( $system, $out, $in->{'qradarVersion'} );

		$out->open_element('interfaces');
		foreach my $interface ( @{ $system->{interfaces}->{interface} } ) {
			_print_interface($interface, $out);
		}
		$out->close_element('interfaces');

		print_if_defined( $system, 'staticRoutes', $out );
		print_if_defined( $system, 'subelementType', $out );

		$out->close_element('subelement');
	}

	$out->close_element('subelements');

	return;
}

# Print the manifest of packet processing for PANOS
sub _print_packet_process_ordering {
	my $out = shift;

	$out->open_element('core:packetProcessingOrdering');

	for my $mPhase ( 'address-translation/destination', 'routing', 'address-translation/source', 'layer2-bridging', 'packet-filtering' ) {
		my ($pPhase, $pPosition) = split /\//, $mPhase;
		$out->open_element('core:packetProcessingElement');
		$out->print_element('core:packetProcessingPhase', $pPhase);
		if( defined $pPosition ) {
			$out->print_element('core:processingPoint', $pPosition);
		}
		$out->close_element('core:packetProcessingElement');
	}
	$out->close_element('core:packetProcessingOrdering');

	return;
}

# Careful, this is also used by the Applications operation
sub parse_applications {
	my ($vsys, $lookup) = @_;

	my $parsed_applications;

	my @applications = $vsys->findnodes('application/entry');
	for my $application ( @applications ) {
		my $parsed_application;
		my $lookup_application;

		my $name = $application->att('name');
		$parsed_application->{'ztApp:appName'} = $name;
		
		my $description = get_value_from_xml('description', $application);
		if ($description) {
			$parsed_application->{'ztApp:appDescription'} = $description;
		}

		my $risk = get_value_from_xml('risk', $application);
		$parsed_application->{'ztApp:appRiskSeverity'} = $risk;

		if(defined($risk)){
			push(@{$lookup_application->{criteria}->{risk}}, $risk);
		}

		# This is a list of known categories used for classification
		my @classifiers = qw(
			category
			subcategory
			technology
			evasive-behavior
			consume-big-bandwidth
			used-by-malware
			able-to-transfer-file
			has-known-vulnerability
			tunnel-other-application
			prone-to-misuse
			pervasive-use
			file-type-ident
			virus-ident
			spyware-ident
			data-ident
			parent-app
		);

		# Each application has a set of classifiers
		foreach my $classifier (@classifiers) {

			my $app_classification;
			$app_classification->{'ztApp:classifierName'} = $classifier;

			my $classifier_value = get_value_from_xml($classifier, $application);
			$app_classification->{'ztApp:classifierValue'} = $classifier_value;

			if(defined($classifier_value)){
				push(@{$lookup_application->{criteria}->{$classifier}}, $classifier_value);
			}

			# Adapter level logging to detect when these strings are longer than the database can handle.
			if (length $app_classification->{'ztApp:classifierName'} > 128) {
				$LOGGER->debug('Clasifier name: ' . $app_classification->{'ztApp:classifierName'} . ' is bigger than 128 characters for application ' . $application->att('name'));
			}
			if (defined $app_classification->{'ztApp:classifierValue'} and length $app_classification->{'ztApp:classifierValue'} > 128) {
				$LOGGER->debug('Clasifier Value: ' . $app_classification->{'ztApp:classifierValue'} . ' is bigger than 128 characters for application ' . $application->att('name'));
			}
			if (defined $app_classification->{'ztApp:classifierValue'}) {
				push @{$parsed_application->{'ztApp:appClassification'}}, $app_classification;
			}
		}


		# Find the protocol/port entries for this application (if any)
		my @ports = $application->findnodes('default/port/member');
		my $lookup_port;
		my $added = 0;
		for my $port_entry (@ports) {
			my $parsed_port;

			# Get the entry and break it up into port and protocol
			my $value = get_value_from_xml('.', $port_entry);
			my ($protocol, $port) = $value =~ /(tcp|udp|icmp|icmp6)\/(\S+)/i;

			if( defined $protocol ) {

				# ICMP is disabled as the Applications schema and client only
				# supports tcp and udp currently
				my $is_icmp = $protocol =~ /icmp/i;

				# Represent a port entry of 'dynamic' as all possible ports
				if ($port =~ /dynamic|any/i) {

					# Application schema can only take one service currently, and does
					# not support ICMP either.
					if( not $is_icmp and not $added ) {
						push @{$parsed_application->{'ztApp:appNetworkPort'}}, {
							'ztApp:protocol' => uc $protocol,
							'ztApp:port' => {
								'ztApp:portRange' => {
									'ztApp:portLow' => 1,
									'ztApp:portHigh' => 65535
								}
							}
						};
						++$added;
					}
					push @{$lookup_port->{portRange}}, {
						portStart => 1,
						portEnd => 65535,
						protocol => lc $protocol
					};
				}else{
					my ($ports, $port_ranges) = unwind_port_expression($port);
					for my $port_value ( @{ $ports } ) {
						if( not $is_icmp and not $added ) {
							push @{$parsed_application->{'ztApp:appNetworkPort'}}, {
								'ztApp:protocol' => uc $protocol,
								'ztApp:port' => {
									'ztApp:portValue' => $port_value
								}
							};
							++$added;
						}
						push @{$lookup_port->{portExpression}}, {
							protocol => lc $protocol,
							port => $port_value,
							operator => 'eq'
						};
					}
					for my $port_range ( @{ $port_ranges } ) {
						my ($low, $high) = split '-', $port_range;

						if( not $is_icmp and not $added ) {
							push @{$parsed_application->{'ztApp:appNetworkPort'}}, {
								'ztApp:protocol' => uc $protocol,
								'ztApp:port' => {
									'ztApp:portRange' => {
										'ztApp:portLow' => $low,
										'ztApp:portHigh' => $high
									}
								}
							};
							++$added;
						}
						push @{$lookup_port->{portRange}}, {
							protocol => lc $protocol,
							portStart => $low,
							portEnd => $high
						};
					}
				}
			}else{
				$LOGGER->debug("WARNING: Unable to add port to application because no protocol was found in '$value'");
				next;
			}
		}

		if($lookup_port){
			$lookup_application->{service_group} = $lookup_port;
		}

		# Store in the lookup table for finding it later
		if($lookup){
			$lookup->add_application($name, $lookup_application);
		}

		push @{ $parsed_applications }, $parsed_application;
	}

	return $parsed_applications;
}

sub parse_predefined_and_shared_applications{
	my ($in) = @_;

	my $lookup = $in->{lookup_tables};
	my $predefined = $in->{show_predefined_applications};
	my $show_config_running = $in->{show_config_running};

	my $predefined_nodes = $predefined->findnodes('/response/result');
	my $predefined_applications = shift(@{$predefined_nodes});
	parse_applications($predefined_applications, $lookup);

	my $shared_nodes = $show_config_running->findnodes('/response/result/config/shared');
	my $shared_applications = shift(@{$shared_nodes});
	parse_applications($shared_applications, $lookup);

	return;
}

=begin to be used for future user group modelling
sub parse_user_groups {
	my $vsys      = shift;
	my $lookup = shift;
	my $show_user_group_name = shift;
	
	my @parsedUserGroups;
	if (!$show_user_group_name){
		my @user_groups = $vsys->findnodes('local-user-database/user-group/entry');
		foreach my $entry (@user_groups){
			my $name = $entry->att('name');
			my $userGroup->{id} = $name;
			# get users
			my @userGroupMembers = $entry->findnodes('user/member');
			foreach my $member (@userGroupMembers){
				my $user = get_value_from_xml( '.', $member );
				push @{$userGroup->{user}}, $user;
			}
			$lookup->add_user_group($name, $userGroup);
			push @parsedUserGroups, $userGroup;
		}
	} else { # must be a vsystem
		foreach my $userGroupName (keys %{$show_user_group_name}){
			my $userGroupResult = $show_user_group_name->{$userGroupName}->findvalue('/response/result');
			my $userGroup;
			foreach my $line (split ("\n", $userGroupResult)){
				if ($line =~ /^short\sname:\s+(.*)/){
					my $userShortName = lc ($1);
					$userGroup->{id} = $userShortName;
					$lookup->add_user_group($userGroupName, $userGroup, $userShortName);
				} elsif ($line =~ /^\[\d+\s+\]\s(.*)/){
					my $user = $1;
					push @{$userGroup->{user}}, lc($user) if $user ne "";
				}
			}
			push @parsedUserGroups, $userGroup if defined $userGroup->{id};
		}
	}
	
	return \@parsedUserGroups;
}
=cut

1;
