package ZipTie::Adapters::PaloAlto::PANOS::Telemetry;

use strict;
use warnings;
use 5.010_000;

use ZipTie::Adapters::PaloAlto::PANOS::AutoLogin qw(get_connection get_response_objects);
use ZipTie::Adapters::PaloAlto::PANOS::LookupTable;
use ZipTie::Adapters::PaloAlto::PANOS::TelemetryParsers qw(parse_system_telemetry parse_interfaces_telemetry parse_mac_telemetry parse_arp_telemetry parse_routes_telemetry print_telemetry);

use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

my $LOGGER = ZipTie::Logger::get_logger();

sub invoke {
	my ($package_name, $telemetry_doc) = @_;

	$LOGGER->debug('Starting telemetry operation.');

	my ($connection_path) = ZipTie::Typer::translate_document( $telemetry_doc, 'connectionPath' );

	my $responses = _get_responses($connection_path);

	#Instantiate the LookupTable object that will be used to pass information between parsers
	$responses->{lookup_tables} = ZipTie::Adapters::PaloAlto::PANOS::LookupTable->new();

	$LOGGER->debug('Opening Filehandle...');
	my $filehandle = get_model_filehandle( 'PANOS', $connection_path->get_ip_address() );

	$LOGGER->debug('Opening Printer...');
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'telemetry' );

	$LOGGER->debug('Parsing Telemetry...');

	$LOGGER->debug('Parsing Supporting Information...');
	parse_system_telemetry( $responses, $printer );

	$LOGGER->debug('Parsing Interfaces...');
	parse_interfaces_telemetry( $responses, $printer );

	$LOGGER->debug('Parsing Mac Table...');
	parse_mac_telemetry( $responses, $printer );

	$LOGGER->debug('Parsing Arp Table...');
	parse_arp_telemetry( $responses, $printer );

	$LOGGER->debug('Parsing Routes...');
	parse_routes_telemetry( $responses, $printer );

	$LOGGER->debug('Successfully Parsed Telemetry Information.');

	$printer->attributes(1);
	$printer->open_discovery_event();

	print_telemetry( $responses, $printer );
	$LOGGER->debug('DiscoveryEvent Document Printed.');

	$printer->close_element('DiscoveryEvent');

	close_model_filehandle($filehandle);
	$LOGGER->debug('Filehandle closed');

	$LOGGER->debug('Telemetry operation complete.');
}

sub _get_responses {
	my ($connection_path) = @_;

        my $connection = get_connection($connection_path);

	my %commands = (
		required => [
			[ show_system_info       => 'type=op&cmd=<show><system><info></info></system></show>' ],
			[ show_interface_all     => 'type=op&cmd=<show><interface>all</interface></show>' ],
			[ show_routing_interface => 'type=op&cmd=<show><routing><interface></interface></routing></show>' ]
		],
		optional => [
			[ show_interface_counters => 'type=op&cmd=<show><counter><interface>all</interface></counter></show>' ],
			[ show_arp_all            => "type=op&cmd=<show><arp><entry name='all'/></arp></show>" ],
			[ show_mac_all            => 'type=op&cmd=<show><mac>all</mac></show>' ],
			[ show_routing_route      => 'type=op&cmd=<show><routing><route></route></routing></show>' ]
		]
	);

	my $responses = get_response_objects($connection, \%commands);

	return $responses;
}

1;
