package ZipTie::Adapters::PaloAlto::PANOS::TelemetryParsers;

use strict;
use warnings;
use 5.010_000;

use XML::Twig::XPath;

use ZipTie::Adapters::PaloAlto::PANOS::Utils qw(get_value_from_xml print_if_defined);
use ZipTie::Adapters::Utils qw(strip_mac get_crep);
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_system_telemetry parse_interfaces_telemetry parse_mac_telemetry parse_arp_telemetry parse_routes_telemetry print_telemetry);

my $LOGGER = ZipTie::Logger::get_logger();

sub parse_system_telemetry {
	my ( $in, $out ) = @_;
	my $show_system_info = $in->{show_system_info};

	$in->{parsed}->{adminIp} = get_value_from_xml( '/response/result/system/ip-address', $show_system_info );
	$in->{parsed}->{sysName} = get_value_from_xml( '/response/result/system/hostname',   $show_system_info );

	return;
}

sub parse_interfaces_telemetry {
	my ( $in, $out ) = @_;
	my $ip_regex = get_crep('cipm');

	my $show_interface_counters = $in->{show_interface_counters};
	my $show_routing_interface  = $in->{show_routing_interface};
	my $show_interface_all      = $in->{show_interface_all};

	my @entries = $show_routing_interface->findnodes('/response/result/interface');
	for my $entry ( @entries ) {

		my $parsed_entry;

		my $vr_name = get_value_from_xml( 'vr', $entry );
		my $if_name = get_value_from_xml( 'name', $entry );

		# If the router name is longer than 13 characters the virtual interface that passes traffic to
		# other virtual routers may have its name truncated. The name for this interface is always
		# the name of the router + '/i3', as far as I know.
		# so check to see if the router name got truncated and fix this.
		if( length $if_name == 16 and $vr_name =~ /$if_name/ ) {
			$if_name = $vr_name . '/i3';
		}
		$parsed_entry->{name} = $if_name;

		# Palo Alto interfaces are named by the system according to their type
		my $type = 'unknown';
		if( $parsed_entry->{name} =~ /ethernet/i ) {
			$type = 'ethernet';
		} elsif( $parsed_entry->{name} =~ /vlan|tunnel|$vr_name/i ) {
			$type = 'other';
		} elsif( $parsed_entry->{name} =~ /loopback/i ) {
			$type = 'softwareLoopback';
		}
		$parsed_entry->{type} = $type;

		$parsed_entry->{operStatus} = 'Down';

		# Status is represented as a '*' character or a blank in the output of 'show routing interfaces'
		my $status = get_value_from_xml( 'status', $entry );
		if( $status =~ /\*/ ) {
			$parsed_entry->{operStatus} = 'Up';
		}

		if( $show_interface_counters ) {
			$parsed_entry->{inputBytes} = get_value_from_xml( '/response/result/ifnet/entry[name="' . $parsed_entry->{name} . '"]/ibytes', $show_interface_counters );
		}
		if( not defined $parsed_entry->{inputBytes} or $parsed_entry->{inputBytes} !~ /\d+/ ) {
			$parsed_entry->{inputBytes} = 0;
		}

		my @addresses = $entry->findnodes('address');
		for my $address_entry (@addresses) {
			my $cidr = get_value_from_xml( '.', $address_entry );
			if( defined $cidr and $cidr =~ /($ip_regex)(\/(\d+))?/ ) {
				my ($address, $mask) = ($1, $3);
				my $ip_entry = { ipAddress => $address, mask => $mask // 32 };
				push @{ $parsed_entry->{ipEntry} }, $ip_entry;
			}
		}

		$in->{lookup_tables}->add_interface( $parsed_entry->{name}, $parsed_entry );
		push @{ $in->{parsed}->{interfaces}->{interface} }, $parsed_entry;
	}

	# Get all the network interfaces
	@entries = $show_interface_all->findnodes('/response/result/ifnet/entry');
	for my $entry (@entries) {

		my $parsed_entry;

		$parsed_entry->{name} = get_value_from_xml( 'name', $entry );

		if( not defined $in->{lookup_tables}->get_interface( $parsed_entry->{name} ) ) {

			# Palo Alto interfaces are named by the system according to their type
			my $type = 'unknown';
			if( $parsed_entry->{name} =~ /ethernet/i ) {
				$type = 'ethernet';
			} elsif( $parsed_entry->{name} =~ /vlan|tunnel/i ) {
				$type = 'other';
			} elsif( $parsed_entry->{name} =~ /loopback/i ) {
				$type = 'softwareLoopback';
			}
			$parsed_entry->{type} = $type;

			# Sub interfaces have a name that is the name of the master interface plus a dot and a number
			my ($basename) = $parsed_entry->{name} =~ m{([^\.]*)(?:\.\d+)?};

			# Use the base name to extract the state of the interface. Assume that if the master is down then all the subs are down
			$parsed_entry->{operStatus} = ucfirst get_value_from_xml( "/response/result/hw/entry[name=\"$basename\"]/state", $show_interface_all );

			# Find this interface in the output of show counters and extract the ibytes entry...if the counter exists (optional command elicited it)
			if( $show_interface_counters ) {
				$parsed_entry->{inputBytes} = get_value_from_xml( '/response/result/ifnet/entry[name="' . $parsed_entry->{name} . '"]/ibytes', $show_interface_counters );
			}
			if( not defined $parsed_entry->{inputBytes} or $parsed_entry->{inputBytes} !~ /\d+/ ) {
				$parsed_entry->{inputBytes} = 0;
			}

			# Extract the first ip address information if it exists
			my $cidr = get_value_from_xml( 'ip', $entry );
			if( defined $cidr and $cidr =~ /($ip_regex)(\/(\d+))?/ ) {
				my ($address, $mask) = ($1, $3);
				my $ip_entry = { ipAddress => $address, mask => $mask // 32 };
				push @{ $parsed_entry->{ipEntry} }, $ip_entry;

				# Interfaces can also have more ip addresses stored in a different place in the XML. Grab them as well
				my @alt_addresses = $entry->findnodes('addr/member');
				for my $alt_address ( @alt_addresses ) {

					$cidr = get_value_from_xml( '.', $alt_address );

					if( defined $cidr and $cidr =~ /($ip_regex)(\/(\d+))?/ ) {
						($address, $mask) = ($1, $3);
						$ip_entry = { ipAddress => $address, mask => $mask // 32 };
						push @{ $parsed_entry->{ipEntry} }, $ip_entry;
					}
				}

			}

			$in->{lookup_tables}->add_interface( $parsed_entry->{name}, $parsed_entry );
			push @{ $in->{parsed}->{interfaces}->{interface} }, $parsed_entry;
		}
	}

	return;
}

sub parse_mac_telemetry {
	my ( $in, $out ) = @_;

	my $show_mac_all = $in->{show_mac_all} or return;  # mac entries were optional as a command

	my @entries = $show_mac_all->findnodes('/response/result/entries/entry');
	for my $entry (@entries) {

		my $mac_address = strip_mac( get_value_from_xml( 'mac', $entry ) );
		my $interface = get_value_from_xml( 'interface', $entry );
		my $vlan = get_value_from_xml( 'vlan', $entry );

		my $parsed_entry = {
			macAddress => $mac_address,
			interface  => $interface,
			vlan       => $vlan
		};

		push @{ $in->{parsed}->{neighbors}->{macTable}->{macEntry} }, $parsed_entry;
	}
	return;
}

sub parse_arp_telemetry {
	my ( $in, $out ) = @_;

	my $show_arp_all = $in->{show_arp_all} or return;  # ARP retrieval was an optional command

	my @entries = $show_arp_all->findnodes('/response/result/entries/entry');
	for my $entry ( @entries ) {

		my $mac_address = strip_mac( get_value_from_xml( 'mac', $entry ) );
		my $interface = get_value_from_xml( 'interface', $entry );
		my $ip_address = get_value_from_xml( 'ip', $entry );

		my $parsed_entry = {
			macAddress => $mac_address,
			interface  => $interface,
			ipAddress  => $ip_address
		};

		push @{ $in->{parsed}->{neighbors}->{arpTable}->{arpEntry} }, $parsed_entry;
	}
	return;
}

sub parse_routes_telemetry {
	my ($in, $out) = @_;
	my $ip_regex = get_crep('cipm');

	my $show_routing_route = $in->{show_routing_route} or return; # Optional

	my @entries = $show_routing_route->findnodes('/response/result/entry');
	for my $entry ( @entries ) {
		my $parsed_entry;

		# Routing protocol is represented by letter flags in the entry
		my $flags = get_value_from_xml('flags', $entry);

		# Default to unknown
		my $protocol = 'UNKNOWN';
		if ($flags =~ /S/){
			####Disabled due to inability of 71 mr2 to handle 'STATIC'
			# $protocol = 'STATIC';
		}elsif($flags =~ /R/){
			$protocol = 'RIP';
		}elsif($flags =~ /O/){
			$protocol = 'OSPF';
		}elsif($flags =~ /B/){
			$protocol = 'BGP';
		}
		$parsed_entry->{protocol} = $protocol;

		# Get the destination IP address that this route applies to
		my $destination = get_value_from_xml('destination', $entry);
		( $parsed_entry->{ipAddress} ) = $destination =~ /($ip_regex)/;

		# In 71_mr2 router ID is optional in the schema, but Ziptie Server breaks with a null pointer exception
		# if it is not included, setting the routerID to be the same as the destination does not make any sense
		# but it is the only value I have. Hopefully it is not actually used in ZT server or QRM core.
		# TODO: We should not cate about this anymore, who is realistically on 7.1.2?
		$parsed_entry->{routerId} = $parsed_entry->{ipAddress};

		# Get the interface that this route exits through
		$parsed_entry->{interface} = get_value_from_xml('interface', $entry);

		# Make sure we don't have an empty attribute if there is no interface
		if( not $parsed_entry->{interface} ) {
			delete $parsed_entry->{interface};
		} elsif( not defined $in->{'lookup_tables'}->get_interface( $parsed_entry->{interface} ) ) {
			$LOGGER->debug("Routing entry for $parsed_entry->{ipAddress} has an unknown interface entry.");
			$LOGGER->debug("Removing interface $parsed_entry->{interface} from the route entry.");
			delete $parsed_entry->{interface};
		}

		push @{ $in->{parsed}->{neighbors}->{routingNeighbors}->{routingNeighbor} }, $parsed_entry;
	}
	return;
}

sub print_telemetry {
	my ( $in, $out ) = @_;

	print_if_defined( $in->{parsed}, 'adminIp', $out );
	print_if_defined( $in->{parsed}, 'sysName', $out );

	$out->attributes(1);
	print_if_defined( $in->{parsed}, 'neighbors', $out );
	print_if_defined( $in->{parsed}, 'interfaces', $out );

	return;
}

1;
