package ZipTie::Adapters::PaloAlto::PANOS::Utils;

use strict;
use warnings;
use 5.010_000;

use XML::Twig::XPath;

use ZipTie::Adapters::Utils qw(parseCIDR get_crep);
use ZipTie::Logger;

our @EXPORT_OK = qw(print_if_defined get_value_from_xml unwind_port_expression parse_cidr_panos get_interface_type_panos convert_rule_action add_interface_grouping);
use Exporter qw(import);

my $LOGGER = ZipTie::Logger::get_logger();

# This function takes a hash, a key and an output printer and prints the SED element identified by the key
# if that element is defined in the hash.
# Inputs:
# $element is a hash of parsed data where the keys are names of SED elements
# $key is the name of a SED element of be printed
# $out is the output printer where the SED element will printed
sub print_if_defined {
	my ( $element, $key, $out ) = @_;

	if ( defined( $element->{$key} ) ) {
		$out->print_element( $key, $element->{$key} );
	}
	return;
}

sub get_value_from_xml {
	my $path = shift;
	my $xml  = shift;

	my $value = $xml->findvalue($path);

	# Adapters common has a perl module it depends upon, perl-libxml-perl-0.08-10.el6.noarch
	# No idea where it comes from or if we NEED it, but our dev machines do not have it
	# When our dev machines execute findvalue it returns a string value
	# When a machine with perl-libxml-perl-0.08-10.el6.noarch installed executes findvalue it
	# returns an XML::XPath::Literal object, the below if-statement handles this environment discrepancy
	if( ref $value eq 'XML::XPath::Literal' ) {
		$value = $value->value();
	}

	if ( $value eq '' ) {
		$value = undef;
	}

	return $value;
}

#given a string containing a port expression, produce a list of ports/portRanges
sub unwind_port_expression {
	my $expression = shift;

	my @ports;
	my @port_ranges;

	#if there is more that one port or range defined they will be separated by commas
	my @list_entries = split ',', $expression;

	for my $entry (@list_entries) {

		#if this item is a range it will be separated by a dash
		if( $entry =~ /-/ ) {
			push @port_ranges, $entry;
		} else {
			push @ports, $entry;
		}
	}

	return \@ports, \@port_ranges;
}

sub parse_cidr_panos {
	my $input = shift;
	my $ip_regex = get_crep('cipm');
	my $cidr;
	
	# If we only have an IP address, just assign it and code a /32.
	# TODO: Handle IPv6 cidrs
	if( $input =~ /^($ip_regex)$/ ) {
		$cidr = { host => $1, prefix => 32 };
	} else {
		$cidr = parseCIDR($input);
	}

	return $cidr;		
}

sub get_interface_type_panos{
	my ($name) = @_;
	my $type = 'unknown';

	if($name =~ /^ethernet/){
		$type = 'ethernet';
	}elsif($name =~ /^ae|vlan/){
		$type = 'virtual';
	}elsif($name =~ /^tunnel/){
		$type = 'tunnel';
	}elsif($name =~ /^loopback/){
		$type = 'softwareLoopback';
	}elsif($name =~ /^hsci/){
		$type = 'serial';
	}elsif($name =~ /^ha/){
		$type = 'other';
	}else{
		$LOGGER->debug("Cannot infer interface type from name '$name', setting to '$type'");
	}

	return $type;
}

sub convert_rule_action {
	my ($rule_action) = @_;

	# Possible PANOS rule actions are:
	#   'allow', 'drop', deny', 'reset-client', 'reset-server' and 'reset-both'
	my $new_rule_action = $rule_action eq 'allow' ? 'permit' : 'deny';

	return $new_rule_action;
}

sub add_interface_grouping {
	my( $interface, $group_name, $group_type ) = @_;

	my $add = 1;
	for my $interface_grouping ( @{ $interface->{interfaceGrouping} } ) {
		if( $interface_grouping->{ifGroupingCharacteristic} eq $group_type
		and $interface_grouping->{ifGroupName} eq $group_name ) {
			$add = 0; last;
		}
	}
	if( $add ) {
		push @{$interface->{interfaceGrouping}}, { ifGroupName => $group_name, ifGroupingCharacteristic => $group_type };
	}
	return;
}

1;
