package ZipTie::Discovery;

use strict;
use warnings;
use 5.010_000;

use English qw( -no_match_vars );    #Friendly names for punctuation variables
use Exporter qw( import );
use Data::Dumper;

use ZipTie::Client;
use ZipTie::ConnectionPath;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::Credentials;
use ZipTie::Recording;
use ZipTie::Logger;

# Set up logging
my $LOGGER = ZipTie::Logger::get_logger();

our @EXPORT_OK = qw( get_credentials get_connection_path get_devices add_or_update_devices );

# Collet the credentials for the management server IP from ziptie
sub get_credentials {
        my ($ziptie_client, $management_server_ip, $possible_protocols) = @_;
        
        # Get credentials, fail if none are found
        my $credentials_service = $ziptie_client->credentials();
        my @credentials = $credentials_service->getCredentialSetsByIpAddress(ipAddress => $management_server_ip);
        if( not scalar @credentials ) {
                _fatal('No credentials found.');
        }

        # Get username and password for the management ip, fail if none are found
        my ($username, $password);
        if( ref($credentials[0]->{credentials}) ne 'ARRAY') {
                $credentials[0]->{credentials} = [$credentials[0]->{credentials}];
        }
        foreach my $credential ( @{$credentials[0]->{credentials}} ) {
                if( $credential->{name} eq 'username' ) {
                        $username = $credential->{value};
                } elsif ( $credential->{name} eq 'password') {
                        $password = $credential->{value};
                }
        }
        if( not $username ) {
                _fatal('Authentication refused with an empty username.');
        }
        if( not $password ) {
                _fatal('Authentication refused with an empty password.');
        }

	# TODO de-https
	# my $chosen_protocol = ${ @{ $possible_protocols } }[0];
	# for my $possible_protocol ( @{ $possible_protocols } ) { }
        # Get port for protocol, if it is enabled, else fail
        my @protocols = $credentials_service->getProtocolConfigByIpAddress(ipAddress => $management_server_ip);
        if( not scalar @protocols ) {
                _fatal('No protocols found.');
        }
        my $port;
        foreach my $protocol ( @{$protocols[0]->{protocols}} ) {
                if( $protocol->{name} eq 'HTTPS' && $protocol->{enabled} eq 'true' ) {
                        $port = $protocol->{port};
                        last;
                }
        }
        if( not $port ) {
                _fatal('Required protocol either disabled or undefined.');
        }

        return($username, $password, $port, 'HTTPS');
}

# Some protocols require a connection path to be build from the collected credentials
sub get_connection_path {
	my ($ziptie_client, $ip_address) = @_;
	my ($username, $password, $port, $protocol) = get_credentials( $ziptie_client, $ip_address );

	my $connection_path = ZipTie::ConnectionPath->new();

	$connection_path->set_credentials( { username => $username, password => $password } );
	$connection_path->set_ip_address($ip_address);

	my $protocol_obj = ZipTie::ConnectionPath::Protocol->new();
	$protocol_obj->set_name($protocol);
	$protocol_obj->set_port($port);
	$connection_path->set_protocols( [ $protocol_obj ] );

	return $connection_path;
}

# Get devices currently in ziptie. Returns a hash where keys are devices' ips
sub get_devices {
	my ($ziptie_client) = @_;
        my %device_hash;

        # Variables for getting page of devices
        my $PAGE_SIZE = 500;
        my $offset = 0;
        my $total;
        my $device_search_service = $ziptie_client->devicesearch();

        # While there are still devices to get
        while( not defined $total or $offset < $total ) {

                # Get devices, 500 at a time from the current offset
                my $page = $device_search_service->search(scheme => 'ipAddress', query => '',   pageData => {pageSize => $PAGE_SIZE, offset => $offset});
                $offset += $PAGE_SIZE;
                $total = $total // $page->{total};

                # Get an array reference of the devices
                my $devices;
                if( $page->{devices} ) {
                        $devices = $page->{devices};
                        if( ref($devices) ne 'ARRAY' ) {
                                $devices = [$devices];
                        }
                }

                # For each device in the array add it to the hash with its ip as the key
                for my $device( @{$devices} ) {
                        $device_hash{$device->{ipAddress}} = $device;
                }
        }
        return \%device_hash;
}

# If the device is already added update it, otherwise add a new device
sub add_or_update_devices {
	my ($ziptie_client, $adapter, $management_server_ip, $existing_devices, $new_devices) = @_;
	my $device_service = $ziptie_client->devices();

	for my $device_ip ( @{ $new_devices } ) {
		$LOGGER->debug("Analyzing device $device_ip...");

		my $device = $existing_devices->{$device_ip};

		# If device details have changed then update device, if it already exists then skip or else if it doesn't exist than add device
		if( $device ) {
			if( not $device->{managementServerIp} or $device->{managementServerIp} ne $management_server_ip ) {
				$device->{managementServerIp} = $management_server_ip;
				$device_service->updateDevice(ipAddress => $device_ip, managedNetwork => 'Default', device => $device);
				$LOGGER->debug("Updated $device_ip node.");
				print "$device_ip,Updated\n";
			} else {
				$LOGGER->debug("Skipping existing device: $device_ip.");
				print "$device_ip,Skipped\n";
			}
		} else {
			my %device = (ipAddress => $device_ip, adapterId => $adapter, managedNetwork => "Default", managementServerIp => $management_server_ip);
			$device_service->createManagedDevice(%device);
			$existing_devices->{$device_ip} = \%device;
			$LOGGER->debug("Adding new device $device_ip.");
			print "$device_ip,Added\n";
		}
	}
	return;
}

sub _fatal {
	my ($msg) = @_;
	print "ERROR: $msg";
	$LOGGER->fatal($msg);
	exit 1;
}

1;
